/**
 * Logback: the reliable, generic, fast and flexible logging framework.
 * Copyright (C) 1999-2009, QOS.ch. All rights reserved.
 *
 * This program and the accompanying materials are dual-licensed under
 * either the terms of the Eclipse Public License v1.0 as published by
 * the Eclipse Foundation
 *
 *   or (per the licensee's choosing)
 *
 * under the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation.
 */
package org.act.xservices.log.classic.util;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.List;

import org.act.xservices.log.classic.BasicConfigurator;
import org.act.xservices.log.classic.LoggerContext;
import org.act.xservices.log.classic.joran.JoranConfigurator;

import org.act.xservices.log.core.joran.spi.JoranException;
import org.act.xservices.log.core.status.ErrorStatus;
import org.act.xservices.log.core.status.InfoStatus;
import org.act.xservices.log.core.status.StatusManager;
import org.act.xservices.log.core.status.WarnStatus;
import org.act.xservices.log.core.util.Loader;
import org.act.xservices.log.core.util.OptionHelper;

// contributors
// Ted Graham, Matt Fowles, see also http://jira.qos.ch/browse/LBCORE-32
/**
 * This class contains log's logic for automatic configuration
 * 
 * @author Ceki Gulcu
 */
public class ContextInitializer {

  final public static String AUTOCONFIG_FILE = "log.xml";
  final public static String TEST_AUTOCONFIG_FILE = "log-test.xml";
  final public static String CONFIG_FILE_PROPERTY = "log.configurationFile";
  final public static String STATUS_LISTENER_CLASS = "log.statusListenerClass";
  final public static String SYSOUT = "SYSOUT";

  final LoggerContext loggerContext;

  public ContextInitializer(LoggerContext loggerContext) {
    this.loggerContext = loggerContext;
  }

  public void configureByResource(URL url) throws JoranException {
    if (url == null) {
      throw new IllegalArgumentException("URL argument cannot be null");
    }
    JoranConfigurator configurator = new JoranConfigurator();
    configurator.setContext(loggerContext);
    configurator.doConfigure(url);
  }

  private URL findConfigFileURLFromSystemProperties(ClassLoader classLoader, boolean updateStatus) {
    String logConfigFile = OptionHelper.getSystemProperty(CONFIG_FILE_PROPERTY);
    
    if (logConfigFile != null) {
      URL result = null;
      try {
        result = new URL(logConfigFile);
        return result;
      } catch (MalformedURLException e) {
        // so, resource is not a URL:
        // attempt to get the resource from the class path
        result = Loader.getResource(logConfigFile, classLoader);
        if (result != null) {
          return result;
        }
        File f = new File(logConfigFile);
        if (f.exists() && f.isFile()) {
          try {
            result = f.toURI().toURL();
            return result;
          } catch (MalformedURLException e1) {
          }
        }
      } finally {
        if (updateStatus) {
          statusOnResourceSearch(logConfigFile, classLoader, result);
        }
      }
    }
    return null;
  }

  public URL findURLOfDefaultConfigurationFile(boolean updateStatus) {
    ClassLoader myClassLoader = Loader.getClassLoaderOfObject(this);
    URL url = findConfigFileURLFromSystemProperties(myClassLoader, updateStatus);
    if (url != null) {
      return url;
    }

    url = Loader.getResource(TEST_AUTOCONFIG_FILE, myClassLoader);
    if (updateStatus) {
      statusOnResourceSearch(TEST_AUTOCONFIG_FILE, myClassLoader, url);
    }
    if (url != null) {
      return url;
    }

    url = Loader.getResource(AUTOCONFIG_FILE, myClassLoader);
    if (updateStatus) {
      statusOnResourceSearch(AUTOCONFIG_FILE, myClassLoader, url);
    }
    return url;
  }

  public void autoConfig() throws JoranException {
	  /**
	   * @author zhuzekun
	   * 2010.7.1
	   */
//    StatusListenerConfigHelper.installIfAsked(loggerContext);
//    URL url = findURLOfDefaultConfigurationFile(true);
//    /**
//     * ļļãconfigureByResource(url)
//     * BasicConfigurator.configure(loggerContext)Ĭ
//     */
//    if (url != null) {
//      //ļ
//      configureByResource(url);
//    } else {
    	/**
    	 * һЩĬϵĻ
    	 */
      BasicConfigurator.configure(loggerContext);
    //}
  }

  private void multiplicityWarning(String resourceName, ClassLoader classLoader ) {
    List<URL> urlList = null;
    StatusManager sm = loggerContext.getStatusManager();
    try {
      urlList = Loader.getResourceOccurenceCount(resourceName, classLoader);
    } catch (IOException e) {
      sm.add(new ErrorStatus("Failed to get url list for resource [" + resourceName + "]",
          loggerContext, e));
    }
    if(urlList != null && urlList.size() > 1) {
      sm.add(new WarnStatus("Resource [" + resourceName + "] occurs multiple times on the classpath.",
          loggerContext));
      for(URL url: urlList) {
      sm.add(new WarnStatus("Resource ["+resourceName+"] occurs at ["+url.toString()+"]",
          loggerContext));
      }
    }
  }
  
  private void statusOnResourceSearch(String resourceName, ClassLoader classLoader, URL url) {
    StatusManager sm = loggerContext.getStatusManager();
    if (url == null) {
      sm.add(new InfoStatus("Could NOT find resource [" + resourceName + "]",
          loggerContext));
    } else {
      sm.add(new InfoStatus("Found resource [" + resourceName + "] at ["+url.toString()+"]",
          loggerContext));
      multiplicityWarning(resourceName, classLoader);
    }
  }

}
