package org.act.xservices.engine.rt.persistence.berkeley;

import java.io.File;
import java.util.ArrayList;

import org.act.xservices.engine.rt.persistence.ActivityInstanceRecord;
import org.act.xservices.engine.rt.persistence.EngineRecord;
import org.act.xservices.engine.rt.persistence.IDatabase;
import org.act.xservices.engine.rt.persistence.ProcessInstanceRecord;

import com.sleepycat.je.DatabaseException;
import com.sleepycat.je.Environment;
import com.sleepycat.je.EnvironmentConfig;
import com.sleepycat.persist.EntityCursor;
import com.sleepycat.persist.EntityStore;
import com.sleepycat.persist.PrimaryIndex;
import com.sleepycat.persist.SecondaryIndex;
import com.sleepycat.persist.StoreConfig;

public class BerkeleyDatabase implements IDatabase {

	private String type = "Berkeley DB";
	
	private Environment env;
    
	private EntityStore store;
	
	private static String envHome = "database";
	
	private static String storeName = "BpelEngine";
	
	private PrimaryIndex<Long,EngineRecord> engineRecordIndex;
	
    private PrimaryIndex<String,ProcessInstanceRecord> processInstanceRecordIndex;
    
    private SecondaryIndex<String,String,ProcessInstanceRecord> processInstanceRecordIndexByType;
    
    private PrimaryIndex<Long,ActivityInstanceRecord> activityInstanceRecordIndex;
    
    private SecondaryIndex<String,Long,ActivityInstanceRecord> activityInstanceRecordIndexByProcessInstanceId;
	
	public BerkeleyDatabase() {
		config();
	}
    
    public String getType() {
    	return type;
    }
    
	public void config() {
		try {
			EnvironmentConfig envConfig = new EnvironmentConfig();
		    envConfig.setReadOnly(false);
		    envConfig.setAllowCreate(true);
		    envConfig.setTransactional(true);
			env = new Environment(new File(envHome), envConfig);
		
		    StoreConfig storeConfig = new StoreConfig();
		    storeConfig.setReadOnly(false);
		    storeConfig.setAllowCreate(true);
		    storeConfig.setTransactional(true);
		    storeConfig.setModel(new EngineEntityModel());
		    store = new EntityStore(env, storeName, storeConfig);
		    
		    engineRecordIndex = store.getPrimaryIndex(Long.class, EngineRecord.class);
		    processInstanceRecordIndex = store.getPrimaryIndex(String.class, ProcessInstanceRecord.class);
		    processInstanceRecordIndexByType = store.getSecondaryIndex(processInstanceRecordIndex, String.class, "type");
		    activityInstanceRecordIndex = store.getPrimaryIndex(Long.class, ActivityInstanceRecord.class);
		    activityInstanceRecordIndexByProcessInstanceId = store.getSecondaryIndex(activityInstanceRecordIndex, String.class, "processInstanceId");
		    
		} catch (DatabaseException e) {
			e.printStackTrace();
		}
	}
	
	public void close() {
		if(store != null) {
			try {
				store.close();
			} catch (DatabaseException e) {
				e.printStackTrace();
			}
		}
		if(env != null) {
			try {
				env.close();
			} catch (DatabaseException e) {
				e.printStackTrace();
			}
		}
	}
	
	public ArrayList<ActivityInstanceRecord> getActivityInstanceRecordsByProcessInstanceId(
			String processInstanceId) {
		ArrayList<ActivityInstanceRecord> activityInstanceRecords = new ArrayList<ActivityInstanceRecord>();
		try {
			EntityCursor<ActivityInstanceRecord> items = activityInstanceRecordIndexByProcessInstanceId.subIndex(processInstanceId).entities();
			try {
				for (ActivityInstanceRecord item : items) {
					activityInstanceRecords.add(item);
				}
		    } finally {
		    	items.close();
		    } 
		} catch (DatabaseException e) {
			e.printStackTrace();
		}
		return activityInstanceRecords;
	}

	public ArrayList<ActivityInstanceRecord> getAllActivityInstanceRecords() {
		ArrayList<ActivityInstanceRecord> activityInstanceRecords = new ArrayList<ActivityInstanceRecord>();
		try {
			EntityCursor<ActivityInstanceRecord> items = activityInstanceRecordIndex.entities();
			try {
				for (ActivityInstanceRecord item : items) {
					activityInstanceRecords.add(item);
				}
		    } finally {
		    	items.close();
		    } 
		} catch (DatabaseException e) {
			e.printStackTrace();
		}
		return activityInstanceRecords;
	}

	public ArrayList<EngineRecord> getAllEngineRecords() {
		ArrayList<EngineRecord> engineRecords = new ArrayList<EngineRecord>();
		try {
			EntityCursor<EngineRecord> items = engineRecordIndex.entities();
			try {
				for (EngineRecord item : items) {
					engineRecords.add(item);
				}
		    } finally {
		    	items.close();
		    } 
		} catch (DatabaseException e) {
			e.printStackTrace();
		}
		return engineRecords;
	}

	public ArrayList<ProcessInstanceRecord> getAllProcessInstanceRecords() {
		ArrayList<ProcessInstanceRecord> processInstanceRecords = new ArrayList<ProcessInstanceRecord>();
		try {
			EntityCursor<ProcessInstanceRecord> items = processInstanceRecordIndex.entities();
			try {
				for (ProcessInstanceRecord item : items) {
					processInstanceRecords.add(item);
				}
		    } finally {
		    	items.close();
		    } 
		} catch (DatabaseException e) {
			e.printStackTrace();
		}
		return processInstanceRecords;
	}

	public ArrayList<ProcessInstanceRecord> getProcessInstanceRecordsByType(
			String type) {
		ArrayList<ProcessInstanceRecord> processInstanceRecords = new ArrayList<ProcessInstanceRecord>();
		try {
			EntityCursor<ProcessInstanceRecord> items = processInstanceRecordIndexByType.subIndex(type).entities();
			try {
				for (ProcessInstanceRecord item : items) {
					processInstanceRecords.add(item);
				}
		    } finally {
		    	items.close();
		    } 
		} catch (DatabaseException e) {
			e.printStackTrace();
		}
		return processInstanceRecords;
	}

	public void insertActivityInstanceRecord(
			ActivityInstanceRecord activityInstanceRecord) {
		try {
			activityInstanceRecordIndex.put(activityInstanceRecord);
		} catch (DatabaseException e) {
			e.printStackTrace();
		}
	}

	public void insertEngineRecord(EngineRecord engineRecord) {
		try {
			engineRecordIndex.put(engineRecord);
		} catch (DatabaseException e) {
			e.printStackTrace();
		}
	}

	public void insertProcessInstanceRecord(
			ProcessInstanceRecord processInstanceRecord) {
		try {
			processInstanceRecordIndex.put(processInstanceRecord);
		} catch (DatabaseException e) {
			e.printStackTrace();
		}
	}

}
