package org.act.xservices.rt.xcontainer.addressing.uitl;

import java.util.List;

import org.act.xservices.rt.xom.Attribute;
import org.act.xservices.rt.xom.Metaset;
import org.act.xservices.rt.xom.Namespace;
import org.act.xservices.rt.xom.Node;
import org.act.xservices.rt.xom.QName;

public class XmlUtils {
	/**
    *
    * @param e1
    * @param e2
    * 
    * @return True, if:
    *         <ul>
    *         <li>The two Elements have the same QName.</li>
    *         <li>They have the same child elements, <b>in the same 
    *         order</b>; this includes Elements and Texts.</li>
    *         <li>They have the same attributes, excluding XML namespace 
    *         and prefix definitions.</li>
    *         </ul>
    *         This method short-circuits to "true" if the two references 
    *         are the same (e1 == e2). It short-circuits to "false" if 
    *         one reference is null and the other isn't.
    * 
    */
   public static boolean equals(Metaset e1, Metaset e2)
   {
       //
       // avoid semantic comparisons if we can
       //
       if (e1 == e2)
           return true;
       
       //
       // they're not the same reference, so if either equals null, 
       // we can quit
       //
       if (e1 == null || e2 == null)
           return false;
       
       //
       // make sure root elements are the same, including tag name
       //
       QName qname1 = e1.getQName();
       QName qname2 = e2.getQName();
       
       if (!qname1.equals(qname2))
           return false;
       
       return haveMatchingChildren(e1, e2) && 
              haveMatchingAttributes(e1, e2);
   }
   
   /**
   *
   * Recursively compares the children in the Element sub-trees to see if 
   * they are of equal name, value, and <b>order</b>. Only Element and 
   * Text nodes are compared.
   * 
   * @param e1
   * @param e2
   * 
   * @return True if the two Elements have the same children, in the same 
   *         order. The comparison only targets Element and Text children, 
   *         not Attr or other DOM Nodes.
   * 
   * @see #haveMatchingAttributes(Element, Element)
   *
   */
  public static boolean haveMatchingChildren(Metaset e1, Metaset e2)
  {
      //
      // compare all the child nodes... (w/o attributes). nodes must be 
      // identical and IN THE SAME ORDER - if this is so, the node types 
      // should match
      //
      
      List<Metaset> children1 = e1.elements();
      List<Metaset> children2 = e2.elements();
      
      int length1 = children1.size();
      int length2 = children2.size();
      
      if (length1 != length2)
          return false;
      
      for (int n = 0; n < length1; ++n)
      {
          Metaset next1 = children1.get(n);
          Metaset next2 = children2.get(n);
          
          short type1 = next1.getNodeType();
          short type2 = next2.getNodeType();
          
          //
          // nodes aren't the same type - WRONG ORDER!
          //
          if (type1 != type2)
              return false;
          
          //
          // only concerned with Text and Element nodes - Comments and 
          // CDATA are ignored
          //
          
          else if (next1.elements().size() == 0)
          {
              String text1 = next1.getTextTrim();
              String text2 = next2.getTextTrim();
              
              if (!text1.equals(text2))
                  return false;
          }
          
          //
          // recurse down the tree to do more comparisons
          //
          else if (!equals(next1, next2))
              return false;
      }
      
      return true;
  }
  
  /**
   * 
   * @param e1
   * @param e2
   * 
   * @return True if the two Elements have the same attributes and 
   *         attribute values. The comparison <b>excludes</b> XML 
   *         namespace declarations, since the prefixes used in an 
   *         XML fragment are not important. 
   *
   */
  public static boolean haveMatchingAttributes(Metaset e1, Metaset e2)
  {
      //
      // the order of the attributes is not important, just the values. 
      // we ignore namespace URI attributes because the prefixes used 
      // in a fragment are also unimportant
      //
      
      List<Attribute> attr1 = e1.attributes();
      List<Attribute> attr2 = e2.attributes();
      
      int length1 = attr1.size();
      int length2 = attr2.size();
      
      List<Attribute> larger = length1 >= length2 ? attr1 : attr2;
      int largerLength = larger.size();
      
      for (int n = 0; n < largerLength; ++n)
      {
          Attribute attr = larger.get(n);
          String name = attr.getName();
          String value = attr.getValue();
          
          //
          // ignore namespace URI declarations...
          //
//          if (!name.startsWith("xmlns:"))
//          {
//        	  Attribute match = larger.getNamedItem(name);
//              
//              if (match == null)
//                  return false;
//              
//              String matchValue = match.getValue();
//              
//              if (matchValue == null || !matchValue.equals(value))
//                  return false;
//          }
      }
      
      return true;
  }
  
  /**
   * 
   * parentMetasetиchildQNameҵӽڵ,ӽڵparentMetasetֱӽڵ;
   * ֽڵлȡϢ(QName)
   *
   * @param xml
   *        The root Element to perform the search on.
   * 
   * @param childQName
   *        The name of the child element whose text is a QName.
   * 
   * @return A fully-resolved QName, or null if there was no text.
   *
   */
  public static QName getQNameFromChildElemText(Metaset parentMetaset, QName childElemQName)
  {
      //
      // we can't just get the text - we need to have the actual 
      // element so we can resolve the QName's namespace
      //
      Metaset child = parentMetaset.element(childElemQName);
      
      if (child == null)
          return null;
      
      String qnameText = child.getTextTrim();
      if(qnameText == null || qnameText.length() == 0)
    	  return null;
      
      int colonIndex = qnameText.indexOf(':');
      String prefix = null;
      if (colonIndex >= 0)
          prefix = qnameText.substring(0, colonIndex);
      
      return parseQName(qnameText, child);
  }
  
  public static QName parseQName(String qname, Metaset namespaceContext)
  {
      int colonIndex = qname.indexOf(':');
      
      String prefix = null;
      
      if (colonIndex >= 0)
          prefix = qname.substring(0, colonIndex);
      
      String localName = qname.substring(colonIndex + 1);
      
      String uri = null;
      Namespace nsQNamespace = null;

      //
      // if possible, try to resolve the prefix to a namespace URI
      //
      if (namespaceContext != null)
    	  nsQNamespace = namespaceContext.getNamespaceForPrefix(prefix);
      
      if(nsQNamespace == null)
    	  return null;
      
      //
      // prefix is not required, but it CANNOT be null
      //
      if (prefix != null && prefix.length() > 0)
          return new QName(nsQNamespace.getURI(), localName, prefix);
      
      return new QName(uri, localName);
  }
  
}
