package org.act.xservices.rt.rpc.client;

import java.lang.reflect.Type;

import org.act.xservices.rt.client.ServiceClient;
import org.act.xservices.rt.client.XServicesException;
import org.act.xservices.rt.xcontainer.rpc.DeserializationException;
import org.act.xservices.rt.xcontainer.rpc.SerializationException;
import org.act.xservices.rt.xcontainer.util.JavaObjectDeserializeUtil;
import org.act.xservices.rt.xcontainer.util.JavaObjectSerializeUtil;
import org.act.xservices.rt.xom.Metaset;
import org.act.xservices.rt.xom.QName;

public class RPCServiceClient extends ServiceClient {

	public RPCServiceClient(String uri) {
		super(uri);
	}

	/**
	 * Return value can be a single a object or an object array (itself an
	 * object) , but it is difficulty to figure the return object correctly
	 * unless we have TyepMapping in the client side too. Until it is finalized
	 * lets return OMElement as return value. And the retuen value will be the
	 * body first element user has to deal with that and create his own object
	 * out of that.
	 * 
	 * @param opName
	 *            Operation QName (to get the body wrapper element)
	 * @param args
	 *            Arraylist of objects
	 * @return Response Metaset
	 * @throws XServicesException
	 *             in case of a problem - this can either be a processing fault
	 *             or a received on-the-wire fault.
	 */
	public Metaset invokeBlocking(QName opName, Object[] args)
			throws XServicesException {
		try {
			Metaset metaset = JavaObjectSerializeUtil.serializeJavaObject(
					opName, args, this.getArgNames(), false);
			return super.sendReceive(metaset);
		} catch (SerializationException e) {
			throw new XServicesException(e);
		}
	}

	/**
	 * @param opName
	 *            Operation QName (to get the body wrapper element)
	 * @param args
	 *            Arraylist of objects
	 * @param returnTypes
	 *            , this array contains the JavaTypes for the return object , it
	 *            could be one or more depending on the return type , most of
	 *            the type array will contain just one element It should be
	 *            noted that the array should only contains JavaTypes NOT real
	 *            object , what this methods does is , get the body first
	 *            element , and if it contains more than one children take it's
	 *            element and convert that to ith javatype and fill the return
	 *            array the array will look like as follows [Integer, String,
	 *            MyBean , etc]
	 * @return Object array , which will contains real object , but the object
	 *         can either be simple type object or the JavaBeans, thats what
	 *         this method can handle right now the return array will contains
	 *         [10, "Axis2Echo", {"foo","baa","11"}]
	 * @throws XServicesException 
	 *             a problem occurred, either locally or on the other side of
	 *             the wire
	 */

	public Object[] invokeBlocking(QName opName, Object[] args,
			Type[] returnTypes) throws XServicesException {
		try {
			Metaset metaset = JavaObjectSerializeUtil.serializeJavaObject(
					opName, args, this.getArgNames(), false);
			Metaset response = super.sendReceive(metaset);
			return JavaObjectDeserializeUtil.deserializeJavaObject(response,
					returnTypes);
		} catch (SerializationException e) {
			throw new XServicesException(e);
		} catch (DeserializationException e) {
			throw new XServicesException(e);
		} 
	}

	/**
	 * Invoke the nonblocking/Asynchronous call
	 * 
	 * @param opName
	 *            Operation QName (to get the body wrapper element)
	 * @param args
	 *            an array of argument Objects
	 * @throws XServicesException 
	 */
	public void invokeNonBlocking(QName opName, Object[] args) throws XServicesException {
		try {
			Metaset metaset = JavaObjectSerializeUtil.serializeJavaObject(
					opName, args, this.getArgNames(), false);
			// call the underlying implementation
			super.sendReceiveNonBlocking(metaset);
		} catch (SerializationException e) {
			throw new XServicesException(e);
		}

	}

	/**
	 * 
	 * @param opName
	 * @param args
	 * @throws XServicesException 
	 */
	public void invokeRobust(QName opName, Object[] args) throws XServicesException {
		Metaset metset;
		try {
			metset = JavaObjectSerializeUtil.serializeJavaObject(opName, args,
					this.getArgNames(), false);
			// call the underline implementation
			super.sendRobust(metset);
		} catch (SerializationException e) {
			throw new XServicesException(e);
		} catch (XServicesException e) {
			throw new XServicesException(e);
		}
	}
}
