package org.act.xservices.rt.xcontainer.rpc.util;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Type;

import org.act.xservices.rt.xcontainer.framework.provider.ProviderProcessException;
import org.act.xservices.rt.xcontainer.framework.provider.ServiceProcessException;
import org.act.xservices.rt.xcontainer.inf.context.IServiceContext;
import org.act.xservices.rt.xcontainer.inf.service.wsdl.WSDL11Constants;
import org.act.xservices.rt.xcontainer.inf.service.wsdl.XcMessage;
import org.act.xservices.rt.xcontainer.inf.service.wsdl.XcMessagePart;
import org.act.xservices.rt.xcontainer.inf.service.wsdl.XcOperation;
import org.act.xservices.rt.xcontainer.rpc.DeserializationException;
import org.act.xservices.rt.xcontainer.util.JavaObjectDeserializeUtil;
import org.act.xservices.rt.xom.Metaset;
import org.act.xservices.rt.xom.Namespace;
import org.act.xservices.rt.xom.QName;

/**
 * 
 * @author viking
 * @date 2010-8-8
 * @since JDK 1.5.0
 * @version 1.0.0
 * 
 *          This class is used to invoke web service, and return the result.
 * 
 * 
 */
public class RPCProcessInvokeUtil {

	/**
	 * Invoke the the web service.
	 * 
	 * @param serviceContext
	 *            Input soap message
	 * @param implClass
	 *            Web service implementation class
	 * @param inonly
	 *            indicate the service whether is inonly mep.
	 * @return invoked result.
	 * @throws ServiceProcessException
	 * @throws ProviderProcessException
	 */
	public static Object invokeBusinessLogic(IServiceContext serviceContext,
			Object implClass, boolean inonly) throws ServiceProcessException,
			ProviderProcessException {
		// get the Web Service implementation class
		Class<?> serviceClass = serviceContext.getXcService().getServiceClass();

		// get the operation of the web service
		XcOperation xcOperation = serviceContext.getXcOperation();
		// get the implementation method
		Method method = findMethod(xcOperation, serviceClass, inonly);
		if (method == null) {
			throw new ProviderProcessException("Method Does Not Exist for "
					+ xcOperation.getQName());
		}
		// get the parameter of the method
		Metaset methodElement = serviceContext.getInMessageBody()
				.getFirstElement();
		if (methodElement == null) {
			throw new ProviderProcessException(
					"Error Parameter in InSOAPMessage for Operation "
							+ xcOperation.getQName());
		}
		// get the input parameter
		XcMessage inputXcMessage = xcOperation.getInputXcMessage();
		try {
			if (inputXcMessage != null) {
				QName paramQName = inputXcMessage.getQName();
				String style = xcOperation.getOperationStyle();
				if (style.equals(WSDL11Constants.STYLE_DOC)) {
					for (XcMessagePart part : inputXcMessage
							.getXcMessagePartMap().values()) {
						paramQName = part.getParameterType();
						break;
					}
				}
				String targetNamespace = serviceContext.getXcService()
						.getTargetNamespace();
				// invoke the web service
				return invokeServiceClass(paramQName, method,
						implClass, methodElement, targetNamespace);			 
			}
			return null;
		} catch (IllegalArgumentException e) {
			throw new ServiceProcessException(e);
		} catch (DeserializationException e) {
			throw new ServiceProcessException(e);
		}
	}

	/**
	 * 
	 * @param xcop
	 * @param implClass
	 * @param inonly
	 * @return
	 * @throws ProviderProcessException
	 */
	private static Method findMethod(XcOperation xcop, Class<?> implClass,
			boolean inonly) throws ProviderProcessException {
		// get the method from the operation if there is.
		Method method = xcop.getMethod();
		if (method != null)
			return method;

		// get the method name
		String methodName = xcop.getQName().getName();
		try {
			// get all methods from the implementation class
			Method[] methods = implClass.getMethods();
			// find the method with the method name of the operation.
			for (int i = 0; i < methods.length; i++) {
				if (methods[i].getName().equals(methodName)) {
					method = methods[i];
					break;
				}
			}
			if (inonly) {
				// the return value must be void.
				if (!method.getReturnType().equals(void.class)) {					
					return null;
				}
			}
			xcop.setMethod(method);
			return method;

		} catch (Exception e) {
			throw new ProviderProcessException("Error: Can not find method: "
					+ methodName + " in class: " + implClass.getName(), e);
		}
	}
	
	/**
	 * 
	 * @param operation
	 * @param method
	 * @param implClass
	 * @param methodElement
	 * @param targetNamespace
	 * @return
	 * @throws DeserializationException
	 * @throws ServiceProcessException
	 */
	private static Object invokeServiceClass(QName operation, Method method,
			Object implClass, Metaset methodElement, String targetNamespace)
			throws DeserializationException, ServiceProcessException {
		try {
			if (operation == null) {
				// method accept empty SOAPbody
				return method.invoke(implClass, new Object[0]);
			} else {
				// get the uri of the in-soap message
				String messageNamespaceURI = operation.getNamespaceURI();
				// get the method's namespace
				Namespace methodNamespace = methodElement.getNamespace();
				if (messageNamespaceURI != null) {
					if (methodNamespace == null) {
						throw new ServiceProcessException(
								"namespace mismatch require "
										+ messageNamespaceURI + " found none");
					}
					if (!messageNamespaceURI.equals(methodNamespace.getURI())) {					 
						if (!targetNamespace.equals(methodNamespace
								.getURI())) {
							throw new ServiceProcessException(
									"namespace mismatch require "
											+ targetNamespace
											+ " found "
											+ methodElement.getNamespace()
													.getURI());
						}
					}
				} else if (methodNamespace != null) {
					throw new ServiceProcessException(
							"namespace mismatch. XServices Operation expects non-namespace "
									+ "qualified element. But received a namespace qualified element");
				}
				// parameters
				Object[] objectArray;
				// get the parameters class of the implementation method
				Type[] parameters = method.getGenericParameterTypes();

				// body parameter is wrapped
				if (methodElement.getParent().elements().size() == 1) {
					objectArray = JavaObjectDeserializeUtil
							.deserializeJavaObject(methodElement, parameters);
				} else {
					objectArray = JavaObjectDeserializeUtil
							.deserializeJavaObject(methodElement.getParent(),
									parameters);
				}
				// invoke the method

				return method.invoke(implClass, objectArray);
			}
		} catch (IllegalArgumentException e) {
			throw new ServiceProcessException(e);
		} catch (IllegalAccessException e) {
			throw new ServiceProcessException(e);
		} catch (InvocationTargetException e) {
			e.printStackTrace();
			throw new ServiceProcessException(e);
		}
	}
}
