package org.act.xservices.rt.xcontainer.util;

import java.lang.reflect.Array;
import java.util.Collection;
import java.util.Map;

import org.act.xservices.rt.metaset.MetasetBuilder;
import org.act.xservices.rt.xcontainer.rpc.SerializationException;
import org.act.xservices.rt.xcontainer.wsdl.typemapping.SimpleTypeMapper;
import org.act.xservices.rt.xcontainer.wsdl.utils.Base64;
import org.act.xservices.rt.xom.Metaset;
import org.act.xservices.rt.xom.Namespace;
import org.act.xservices.rt.xom.QName;

/**
 * 
 * @author viking
 * @date 2010-7-15
 * @since JDK 1.5.0
 * @version 1.0.0
 * 
 *          This class is used to serialize Object to Metaset. The object type
 *          can be simple type, array,collection and javaBean.
 * 
 */
public class JavaObjectSerializeUtil {
	/**
	 * MetasetBuilder instance
	 */
	private static MetasetBuilder builder = MetasetBuilder.getInstance();

	/**
	 * Serialize operation to metaset.The object can be simple
	 * type,array,arraylist,javabean.
	 * 
	 * @param opQName
	 *            Operation's QName which is used to wrap parameters
	 * @param parameters
	 *            all Parameters of Operation
	 * @param qualified
	 *            where add the xsdNamespace to each parameter.
	 * @return
	 * @throws SerializationException
	 */
	public static Metaset serializeJavaObject(QName opQName,
			Object[] parameters, String[] argNames, boolean qualified)
			throws SerializationException {

		// wrapped operation
		Metaset operation = getItemMetaset(opQName.getName(),opQName.getNamespace(),true);
		Namespace itemNS = null;
		if (qualified) {
			itemNS = opQName.getNamespace();
		}
		// parameter count
		int parameterCount = 0;
		if (parameters != null) {
			for (Object parameter : parameters) {
				// arg name
				String itemName = null;
				if (argNames != null && parameterCount < argNames.length) {
					itemName = argNames[parameterCount++];
				} else {
					itemName = SerializeConstants.ITEM_NAME + parameterCount++;
				}
				// serialize each parameter
				Metaset[] args = serializeParameter(parameter, itemName,
						itemNS, qualified);
				for (Metaset arg : args) {
					// add serialized parameter
					operation.add(arg);
				}
			}
		}

		// 
		return operation;
	}

	public static Metaset[] serializeParameter(Object parameter,
			String itemName, Namespace itemNS, boolean qualified)
			throws SerializationException {
		if (parameter == null) {

			// null parameter
			Metaset result = serializeNull(itemName, itemNS, qualified);
			return new Metaset[] { result };
		} else if (SimpleTypeMapper.isSimpleType(parameter)) {

			// simple type parameter
			Metaset result = serializeSimpleType(parameter, itemName, itemNS,
					qualified);
			return new Metaset[] { result };
		} else if (parameter instanceof byte[]) {

			// byte[]
			Metaset result = serializeByteArray(parameter, itemName, itemNS,
					qualified);
			return new Metaset[] { result };
		} else if (parameter.getClass().isArray()) {

			// array parameter
			return serializeArray(parameter, itemName, itemNS, qualified);
		} else if (SimpleTypeMapper.isCollection(parameter.getClass())) {
			// collection parameter
			Metaset result = serializeCollection(parameter, itemName, itemNS,
					qualified);
			return new Metaset[] { result };
		} else if (parameter instanceof Metaset) {

			// metaset parameter
			Metaset result = serializeMetaset(parameter, itemName, itemNS,
					qualified);
			return new Metaset[] { result };
		} else if (SimpleTypeMapper.isMap(parameter.getClass())) {
			// map
			Metaset result = serializeMap(parameter, itemName, itemNS,
					qualified);
			return new Metaset[] { result };

		} else if (SimpleTypeMapper.isDataHandler(parameter.getClass())) {

			// data handler
			Metaset result = serializeDatahandler(parameter, itemName, itemNS,
					qualified);
			return new Metaset[] { result };
		}

		// java bean
		Metaset result = JavaBeanUtil.serializeJavaBean(parameter, itemNS,
				qualified);
		return new Metaset[] { result };

	}

	/**
	 * Process null parameter, if partName is null then return <itemNum
	 * xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:nil="true"/>
	 * else return <t:p xmlns:t="xx"
	 * xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:nil="true"/>
	 * where partName is QName("xx","p","t");
	 * 
	 * @param partName
	 *            arg item's QName
	 * @param parameterCount
	 *            item count
	 * @return Metaset
	 */
	private static Metaset serializeNull(String itemName, Namespace itemNS,
			boolean qualified) {
		Metaset nullItem = getItemMetaset(itemName, itemNS, qualified);
		// add attribute
		nullItem.addAttribute(SerializeConstants.NIL_QNAME,
				SerializeConstants.TRUE);
		return nullItem;
	}

	/**
	 * Process simple type parameter, like int ,String , long and so on. All the
	 * simple type are serialized to String value.
	 * 
	 * @param parameter
	 *            parameter value
	 * @param partName
	 *            arg item's QName
	 * @param parameterCount
	 *            item count 0 * @return Metaset
	 */
	private static Metaset serializeSimpleType(Object parameter,
			String itemName, Namespace itemNS, boolean qualified) {
		Metaset item = getItemMetaset(itemName, itemNS, qualified);
		item.setText(SimpleTypeMapper.getStringValue(parameter));
		return item;
	}

	/**
	 * Process array type, like int[],Integer[], Object[].... are allowed,
	 * exception byte[]
	 * 
	 * @param parameter
	 *            parameter value
	 * @param partName
	 *            arg item's QName
	 * @param qualifed
	 *            indicate that the item whether add namespace
	 * @param parameterCount
	 *            item count
	 * @return Metaset
	 * @throws SerializationException
	 */
	private static Metaset[] serializeArray(Object parameter, String itemName,
			Namespace itemNS, boolean qualified) throws SerializationException {
		int length = Array.getLength(parameter);
		Metaset[] array = new Metaset[length];

		for (int itemCount = 0; itemCount < length; itemCount++) {
			Object item = Array.get(parameter, itemCount);

			Metaset[] args = serializeParameter(item, itemName, itemNS,
					qualified);
			array[itemCount] = args[0];
		}
		return array;
	}

	/**
	 * Encode byte[] to base64
	 * 
	 * @param parameter
	 *            parameter value
	 * @param partName
	 *            arg item's QName
	 * @param parameterCount
	 *            item count
	 * @return Metaset
	 */
	private static Metaset serializeByteArray(Object parameter,
			String itemName, Namespace itemNS, boolean qualified) {
		Metaset item = getItemMetaset(itemName, itemNS, qualified);
		String base64 = Base64.encode((byte[]) parameter);
		item.setText(base64);
		return item;
	}

	/**
	 * Process DataHandler
	 * 
	 * @param parameter
	 *            parameter value
	 * @param partName
	 *            arg item's QName
	 * @param parameterCount
	 *            item count
	 * @return Metaset
	 */
	private static Metaset serializeDatahandler(Object parameter,
			String itemName, Namespace itemNS, boolean qualified) {
		Metaset item = getItemMetaset(itemName, itemNS, qualified);
		item.setText(parameter.toString());
		return item;
	}

	/**
	 * Process Metaset type, just only wrap it.
	 * 
	 * @param parameter
	 *            parameter value
	 * @param partName
	 *            arg item's QName
	 * @param parameterCount
	 *            item count
	 * @return Metaset
	 */
	private static Metaset serializeMetaset(Object parameter, String itemName,
			Namespace itemNS, boolean qualified) {
		Metaset item = getItemMetaset(itemName, itemNS, qualified);
		Metaset result = (Metaset) parameter;
		result.detach();
		item.add(result);
		return item;
	}

	/**
	 * Process collection, change collection to array , and then process it.
	 * 
	 * @param parameter
	 *            parameter value
	 * @param partName
	 *            arg item's QName
	 * @param qualifed
	 *            indicate that the item whether add namespace
	 * @param parameterCount
	 *            item count
	 * @return Metaset
	 * @throws SerializationException
	 */
	private static Metaset serializeCollection(Object parameter,
			String itemName, Namespace itemNS, boolean qualified)
			throws SerializationException {
		Metaset arrayList = getItemMetaset(itemName, itemNS, qualified);
		Collection<?> collection = (Collection<?>) parameter;
		int itemCount = 0;
		for (Object item : collection) {

			String name = SerializeConstants.ITEM_NAME + itemCount++;
			Metaset[] items = serializeParameter(item, name, itemNS, qualified);
			for (Metaset itemMetaset : items) {
				arrayList.add(itemMetaset);
			}
		}
		return arrayList;
	}

	/**
	 * Serialize Map
	 * 
	 * @param parameter
	 * @param partName
	 * @param parameterCount
	 * @param qualified
	 * @return
	 * @throws SerializationException
	 */
	private static Metaset serializeMap(Object parameter, String itemName,
			Namespace itemNS, boolean qualified) throws SerializationException {
		Metaset mapTable = getItemMetaset(itemName, itemNS, qualified);
		Map<?, ?> map = (Map<?, ?>) parameter;
		// int itemCount = 0;
		for (Object key : map.keySet()) {
			Metaset[] keyMetasets = serializeParameter(key,
					SerializeConstants.KEY, itemNS, qualified);
			for (Metaset keyMetaset : keyMetasets) {
				mapTable.add(keyMetaset);
			}

			Object value = map.get(key);
			Metaset[] valueMetasets = serializeParameter(value,
					SerializeConstants.VALUE, itemNS, qualified);
			for (Metaset valueMetaset : valueMetasets) {
				mapTable.add(valueMetaset);
			}
		}
		return mapTable;
	}

	/**
	 * Create a new Metaset Element.
	 * 
	 * @param partName
	 *            item' QName
	 * @param parameterCount
	 *            item count
	 * @return Metaset
	 */
	public static Metaset getItemMetaset(String partName, Namespace partNS,
			boolean qualified) {
		Metaset item = null;
		if (qualified) {
			Namespace ns = null;
			if (partNS != null) {
				String prefix = partNS.getPrefix();
				if (prefix == null || prefix.equals("")) {
					prefix = SerializeConstants.PREFIX;
					ns = new Namespace(prefix, partNS.getURI());
				} else {
					ns = partNS;
				}
			}
			item = builder.newMetaset(new QName(partName, ns));
		} else {
			item = builder.newMetaset(partName);
		}
		return item;
	}

}
