package org.act.xservices.rt.soap.builder;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.Reader;
import java.net.URL;

import org.act.xservices.rt.soap.ISOAPMessage;
import org.act.xservices.rt.soap.exception.SOAPException;
import org.act.xservices.rt.soap.factory.SOAPFactory;
import org.act.xservices.rt.xom.DocumentException;
import org.act.xservices.rt.xom.DocumentFactory;
import org.act.xservices.rt.xom.io.SAXReader;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;

/**
 * 
 * @author Viking.Liu
 * @date 2010-6-5
 * @since JDK 1.5
 * @version 1.0.0
 * 
 */
/**
 * SOAPSAXBuilder use SAX to parse the SOAP Message.  
 */
public class SOAPSAXBuilder extends SAXReader implements ISOAPBuilder{

	/** SOAP Factory */
	private SOAPFactory factory;

	public SOAPSAXBuilder() {
		super();
	}

	public SOAPSAXBuilder(boolean validating) {
		super(validating);
	}

	public SOAPSAXBuilder(SOAPFactory factory, boolean validating) {
		super(factory, validating);
		this.factory = factory;
	}

	public SOAPSAXBuilder(SOAPFactory factory) {
		super(factory);
		this.factory = factory;
	}

	public SOAPSAXBuilder(String xmlReaderClassName, boolean validating)
			throws SAXException {
		super(xmlReaderClassName, validating);
	}

	public SOAPSAXBuilder(String xmlReaderClassName) throws SAXException {
		super(xmlReaderClassName);
	}

	public SOAPSAXBuilder(XMLReader xmlReader, boolean validating) {
		super(xmlReader, validating);
	}

	public SOAPSAXBuilder(XMLReader xmlReader) {
		super(xmlReader);
	}

	@Override
	public ISOAPMessage read(File file) throws DocumentException {
		try {
			/*
			 * We cannot convert the file to an URL because if the filename
			 * contains '#' characters, there will be problems with the URL in
			 * the InputSource (because a URL like
			 * http://myhost.com/index#anchor is treated the same as
			 * http://myhost.com/index) Thanks to Christian Oetterli
			 */
			InputSource source = new InputSource(new FileInputStream(file));
			if (this.getEncoding() != null) {
				source.setEncoding(this.getEncoding());
			}
			String path = file.getAbsolutePath();

			if (path != null) {
				// Code taken from Ant FileUtils
				StringBuffer sb = new StringBuffer("file://");

				// add an extra slash for filesystems with drive-specifiers
				if (!path.startsWith(File.separator)) {
					sb.append("/");
				}

				path = path.replace('\\', '/');
				sb.append(path);

				source.setSystemId(sb.toString());
			}

			return read(source);
		} catch (FileNotFoundException e) {
			throw new DocumentException(e.getMessage(), e);
		}
	}

	@Override
	public ISOAPMessage read(InputStream in, String systemId)
			throws DocumentException {
		InputSource source = new InputSource(in);
		source.setSystemId(systemId);
		if (this.getEncoding() != null) {
			source.setEncoding(this.getEncoding());
		}
		return read(source);
	}

	@Override
	public ISOAPMessage read(InputStream in) throws DocumentException {
		InputSource source = new InputSource(in);
		if (this.getEncoding() != null) {
			source.setEncoding(this.getEncoding());
		}

		return read(source);
	}

	@Override
	public ISOAPMessage read(Reader reader, String systemId)
			throws DocumentException {
		InputSource source = new InputSource(reader);
		source.setSystemId(systemId);
		if (this.getEncoding() != null) {
			source.setEncoding(this.getEncoding());
		}

		return read(source);
	}

	@Override
	public ISOAPMessage read(Reader reader) throws DocumentException {
		InputSource source = new InputSource(reader);
		if (this.getEncoding() != null) {
			source.setEncoding(this.getEncoding());
		}

		return read(source);
	}

	@Override
	public ISOAPMessage read(String systemId) throws DocumentException {
		InputSource source = new InputSource(systemId);
		if (this.getEncoding() != null) {
			source.setEncoding(this.getEncoding());
		}

		return read(source);
	}

	@Override
	public ISOAPMessage read(URL url) throws DocumentException {
		String systemID = url.toExternalForm();
		InputSource source = new InputSource(systemID);
		if (this.getEncoding() != null) {
			source.setEncoding(this.getEncoding());
		}
		return read(source);
	}

	@Override
	public ISOAPMessage read(InputSource in) throws DocumentException {
		try {
			XMLReader reader = getXMLReader();

			reader = installXMLFilter(reader);

			EntityResolver thatEntityResolver = this.getEntityResolver();

			if (thatEntityResolver == null) {
				thatEntityResolver = createDefaultEntityResolver(in
						.getSystemId());
				this.setEntityResolver(thatEntityResolver);
			}

			reader.setEntityResolver(thatEntityResolver);

			SOAPSAXContentHandler contentHandler = new SOAPSAXContentHandler(
					this.factory, this.getDispatchHandler());
			contentHandler.setEntityResolver(thatEntityResolver);
			contentHandler.setInputSource(in);

			boolean internal = isIncludeInternalDTDDeclarations();
			boolean external = isIncludeExternalDTDDeclarations();

			contentHandler.setIncludeInternalDTDDeclarations(internal);
			contentHandler.setIncludeExternalDTDDeclarations(external);
			contentHandler.setMergeAdjacentText(isMergeAdjacentText());
			contentHandler.setStripWhitespaceText(isStripWhitespaceText());
			contentHandler.setIgnoreComments(isIgnoreComments());
			reader.setContentHandler(contentHandler);

			configureReader(reader, contentHandler);

			reader.parse(in);

			return contentHandler.getSOAPMessage();
		} catch (Exception e) {
			if (e instanceof SAXParseException) {
				// e.printStackTrace();
				SAXParseException parseException = (SAXParseException) e;
				String systemId = parseException.getSystemId();

				if (systemId == null) {
					systemId = "";
				}

				String message = "Error on line "
						+ parseException.getLineNumber() + " of document "
						+ systemId + " : " + parseException.getMessage();

				throw new DocumentException(message, e);
			} else {
				throw new DocumentException(e.getMessage(), e);
			}
		}
	}

	/**
	 * @param documentFactory
	 *            is must be SOAPFactory,or there will occur errors.
	 */
	@Override
	public void setDocumentFactory(DocumentFactory documentFactory) {
		super.setDocumentFactory(documentFactory);
		if (documentFactory instanceof SOAPFactory) {
			this.factory = (SOAPFactory) documentFactory;
		}
	}

	public void setSOAPFactory(SOAPFactory soapFactory) {
		super.setDocumentFactory(soapFactory);
		this.factory = soapFactory;
	}

	public ISOAPMessage readSOAPMessage(InputStream in) throws SOAPException {
		try {
			return this.read(in);
		} catch (DocumentException e) {
			throw new SOAPException(e.getMessage());
		}
	}

	public ISOAPMessage readSOAPMessage(String filepath) throws SOAPException {
		
		try {
			return this.read(new File(filepath));
		} catch (DocumentException e) {
			throw new SOAPException(e.getMessage());
		}
	}

	public ISOAPMessage readSOAPMessage(Reader reader) throws SOAPException {		
		try {
			return this.read(reader);
		} catch (DocumentException e) {
			throw new SOAPException(e.getMessage());
		}
		
	}
}
