package org.act.xservices.rt.soap.builder;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.Reader;
import java.util.Iterator;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.events.Attribute;
import javax.xml.stream.events.Namespace;
import javax.xml.stream.events.StartDocument;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;

import org.act.xservices.rt.soap.ISOAPHeader;
import org.act.xservices.rt.soap.ISOAPMessage;
import org.act.xservices.rt.soap.constants.SOAP11Constants;
import org.act.xservices.rt.soap.constants.SOAP12Constants;
import org.act.xservices.rt.soap.exception.SOAPException;
import org.act.xservices.rt.soap.exception.SOAPProcException;
import org.act.xservices.rt.soap.factory.SOAPFactory;
import org.act.xservices.rt.soap.impl.SOAPMessageImpl;
import org.act.xservices.rt.soap.util.SOAPBuilderhelper;
import org.act.xservices.rt.xom.Metaset;
import org.act.xservices.rt.xom.Node;
import org.act.xservices.rt.xom.io.STAXEventReader;

import com.bea.xml.stream.MXParserFactory;

/**
 * 
 * @author Viking.Liu
 * @date 2010-6-5
 * @since JDK 1.5
 * @version 1.0.0
 * 
 */
/**
 * SOAPStAXBuilder use StAX to parse the SOAP Message.
 */
public class SOAPStAXBuilder extends STAXEventReader implements ISOAPBuilder {

	private SOAPFactory soapFactory;
	private XMLInputFactory inputFactory = new MXParserFactory();// XMLInputFactory.newInstance();

	public SOAPStAXBuilder() {
		super();
	}

	public SOAPStAXBuilder(SOAPFactory factory) {
		this.soapFactory = factory;
		this.setDocumentFactory(factory);
	}

	public ISOAPMessage readSOAPMessage(InputStream in) {
		try {
			return this.readDocument(in);
		} catch (XMLStreamException e) {
			throw new SOAPException(e);
		}
	}

	public ISOAPMessage readSOAPMessage(String filepath) {
		try {
			return this
					.readSOAPMessage(new FileInputStream(new File(filepath)));
		} catch (FileNotFoundException e) {
			throw new SOAPException(e);
		}
	}

	public ISOAPMessage readSOAPMessage(Reader reader) {
		try {
			return this.readDocument(reader);
		} catch (XMLStreamException e) {
			throw new SOAPException(e);
		}
	}

	@Override
	public ISOAPMessage readDocument(InputStream is, String systemId)
			throws XMLStreamException {
		XMLEventReader eventReader = inputFactory.createXMLEventReader(
				systemId, is);
		try {
			return readDocument(eventReader);
		} finally {
			eventReader.close();
		}
	}

	@Override
	public ISOAPMessage readDocument(InputStream is) throws XMLStreamException {
		return readDocument(is, null);
	}

	@Override
	public ISOAPMessage readDocument(Reader reader, String systemId)
			throws XMLStreamException {
		XMLEventReader eventReader = inputFactory.createXMLEventReader(
				systemId, reader);
		try {
			return readDocument(eventReader);
		} finally {
			eventReader.close();
		}
	}

	@Override
	public ISOAPMessage readDocument(XMLStreamReader reader)
			throws XMLStreamException {
		XMLEventReader eventReader = inputFactory.createXMLEventReader(reader);
		try {
			return readDocument(eventReader);
		} finally {
			eventReader.close();
		}
	}

	@Override
	public ISOAPMessage readDocument(Reader reader) throws XMLStreamException {
		return readDocument(reader, null);
	}

	@Override
	public ISOAPMessage readDocument(XMLEventReader reader)
			throws XMLStreamException {
		ISOAPMessage message = null;
		while (reader.hasNext()) {
			XMLEvent nextEvent = reader.peek();
			int type = nextEvent.getEventType();
			switch (type) {
			case XMLStreamConstants.START_DOCUMENT:
				StartDocument event = (StartDocument) reader.nextEvent();
				if (message == null) {
					// create SOAPMessage
					if (event.encodingSet()) {
						String encodingScheme = event
								.getCharacterEncodingScheme();
						message = new SOAPMessageImpl(encodingScheme);
					} else {
						message = new SOAPMessageImpl();
					}
				} else {
					// duplicate or misplaced xml declaration
					String msg = "Unexpected StartDocument event";
					throw new XMLStreamException(msg, event.getLocation());
				}
				break;

			case XMLStreamConstants.END_DOCUMENT:
			case XMLStreamConstants.SPACE:
			case XMLStreamConstants.CHARACTERS:
				// skip end SOAPMessage and space outside the root element
				reader.nextEvent();
				break;

			default:
				if (message == null) {
					// create SOAPMessage
					message = soapFactory.createSOAPMessage();
				}
				Node n = readNode(reader);
				message.add(n);
			}
		}
		return message;
	}

	@SuppressWarnings("unchecked")
	@Override
	public Metaset createElement(StartElement startEvent) {
		QName qname = startEvent.getName();
		org.act.xservices.rt.xom.QName elemName = createQName(qname);

		try {
			// Create a soap Factory
			if (elemName.getNamespaceURI().equals(
					SOAP11Constants.SOAP_ENVELOPE_NAMESPACE_URI)) {
				soapFactory = SOAPFactory.getSOAP11Factory();
			} else if (elemName.getNamespaceURI().equals(
					SOAP12Constants.SOAP_ENVELOPE_NAMESPACE_URI)) {
				soapFactory = SOAPFactory.getSOAP12Factory();
			}
			Metaset elem = null;
			if (headerDeep == 2) {
				elem = soapFactory.createSOAPHeaderBlock(elemName);
			} else {
				// Create a SOAP Element
				elem = SOAPBuilderhelper.createElement(elemName, soapFactory);
			}

			if (elem instanceof ISOAPHeader) {
				headerDeep = 1;
			}

			// create attributes
			for (Iterator i = startEvent.getAttributes(); i.hasNext();) {
				Attribute attr = (Attribute) i.next();
				elem.addAttribute(createQName(attr.getName()), attr.getValue());
			}

			// create namespaces
			for (Iterator i = startEvent.getNamespaces(); i.hasNext();) {
				Namespace ns = (Namespace) i.next();
				elem.addNamespace(ns.getPrefix(), ns.getNamespaceURI());
			}
			return elem;
		} catch (SOAPProcException e) {
			e.printStackTrace();
		}
		return null;
	}

	public void setSOAPFactory(SOAPFactory soapFactory) {
		if (soapFactory != null) {
			this.soapFactory = soapFactory;
		} else {
			this.soapFactory = SOAPFactory.getInstance();
			this.setDocumentFactory(soapFactory);
		}
	}
}
