package org.act.xservices.rt.soap.impl;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.act.xservices.rt.soap.IRolePlayer;
import org.act.xservices.rt.soap.ISOAPHeader;
import org.act.xservices.rt.soap.ISOAPHeaderBlock;
import org.act.xservices.rt.soap.ISOAPVersion;
import org.act.xservices.rt.soap.constants.SOAP12Constants;
import org.act.xservices.rt.soap.constants.SOAP12Version;
import org.act.xservices.rt.soap.constants.SOAPConstants;
import org.act.xservices.rt.soap.exception.SOAPException;
import org.act.xservices.rt.soap.exception.SOAPProcException;
import org.act.xservices.rt.xom.Metaset;
import org.act.xservices.rt.xom.Namespace;
import org.act.xservices.rt.xom.Node;
import org.act.xservices.rt.xom.NodeHelper;
import org.act.xservices.rt.xom.QName;
import org.act.xservices.rt.xom.tree.BackedList;
import org.act.xservices.rt.xom.tree.DefaultElement;
/**
 * 
 * @author Viking.Liu
 * @date 2010-6-3
 * @since JDK 1.5
 * @version 1.0.0
 * 
 */
/**
 * A local interface we can use to make "header checker" objects which can be
 * used by HeaderIterators to filter results. This really SHOULD be done with
 * anonymous classes:
 * <p/>
 * public void getHeadersByRole(final String role) { return new HeaderIterator()
 * { public boolean checkHeader(SOAPHeaderBlock header) { ... if
 * (role.equals(headerRole)) return true; return false; } } }
 * <p/>
 * ...but there appears to be some kind of weird problem with the JVM not
 * correctly scoping the passed "role" value in a situation like the above. As
 * such, we have to make Checker objects instead (sigh).
 */
interface Checker {
	boolean checkHeader(ISOAPHeaderBlock header);
}

/**
 * A Checker to make sure headers match a given role. If the role we're looking
 * for is null, then everything matches.
 */
class RoleChecker implements Checker {
	String role;

	public RoleChecker(String role) {
		this.role = role;
	}

	public boolean checkHeader(ISOAPHeaderBlock header) {
		if (role == null) {
			return true;
		}
		String thisRole = header.getRole();
		return (role.equals(thisRole));
	}
}

/**
 * This Checker uses a RolePlayer to return the appropriate headers for that
 * RolePlayer to process. Ignore "none", always "next", etc.
 */
class RolePlayerChecker implements Checker {
	IRolePlayer rolePlayer;

	/** Optional namespace - if non-null we'll only return headers that match */
	String namespace;

	/**
	 * Constructor.
	 * 
	 * @param rolePlayer
	 *            the RolePlayer to check against. This can be null, in which
	 *            case we assume we're the ultimate destination.
	 */
	public RolePlayerChecker(IRolePlayer rolePlayer) {
		this.rolePlayer = rolePlayer;
	}

	public RolePlayerChecker(IRolePlayer rolePlayer, String namespace) {
		this.rolePlayer = rolePlayer;
		this.namespace = namespace;
	}

	public boolean checkHeader(ISOAPHeaderBlock header) {
		// If we're filtering on namespace, check that first since the compare
		// is simpler.
		if (namespace != null) {
			Namespace headerNamespace = header.getNamespace();
			if (headerNamespace == null
					|| !namespace.equals(headerNamespace.getURI())) {
				return false;
			}
		}

		String role = header.getRole();
		ISOAPVersion version = header.getVersion();

		// 1. If role is ultimatedest, go by what the rolePlayer says
		if (role == null
				|| role.equals("")
				|| (version instanceof SOAP12Version && role
						.equals(SOAP12Constants.SOAP_ROLE_ULTIMATE_RECEIVER))) {
			return (rolePlayer == null || rolePlayer.isUltimateDestination());
		}

		// 2. If role is next, always return true
		if (role.equals(version.getNextRoleURI()))
			return true;

		// 3. If role is none, always return false
		if (version instanceof SOAP12Version
				&& role.equals(SOAP12Constants.SOAP_ROLE_NONE)) {
			return false;
		}

		// 4. Return t/f depending on match
		List<String> roles = (rolePlayer == null) ? null : rolePlayer
				.getRoles();
		if (roles != null) {
			for (Iterator<String> i = roles.iterator(); i.hasNext();) {
				String thisRole = (String) i.next();
				if (thisRole.equals(role)){
					return true;
				}
			}
		}
		return false;
	}
}

/** A Checker to see that we both match a given role AND are mustUnderstand=true */
class MURoleChecker extends RoleChecker {
	public MURoleChecker(String role) {
		super(role);
	}

	public boolean checkHeader(ISOAPHeaderBlock header) {
		try {
			if (header.getMustUnderstand())
				return super.checkHeader(header);
		} catch (SOAPProcException e) {
			e.printStackTrace();
		}
		return false;
	}
}

public abstract class SOAPHeaderImpl extends DefaultElement implements
		ISOAPHeader {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * An Iterator which walks the header list as needed, potentially filtering
	 * as we traverse.
	 */
	class HeaderIterator implements Iterator<ISOAPHeaderBlock> {
		ISOAPHeaderBlock current;
		boolean advance = false;
		boolean detach = false;
		Checker checker;

		private int index = 0;

		public HeaderIterator() {
			this(null);
		}

		public HeaderIterator(Checker checker) {
			this.checker = checker;
		}

		public HeaderIterator(Checker checker, boolean detach) {
			this.checker = checker;
			this.detach = detach;
		}

		public void remove() {
		}

		public boolean checkHeader(ISOAPHeaderBlock header) {
			if (checker == null)
				return true;
			return checker.checkHeader(header);
		}

		public boolean hasNext() {
			if (advance) {
				return current != null;
			}
			advance = true;
			current = getNextSOAPHeaderBlock();
			if (current != null) {
				return true;
			}
			return false;
		}

		public ISOAPHeaderBlock next() {			
			
			if(!advance){
				hasNext();
			}
			ISOAPHeaderBlock ret = current;
			if (ret != null) {
				advance = false;
			}
			if (detach) {
				ret.detach();
			}
			return ret;
		}

		private ISOAPHeaderBlock getNextSOAPHeaderBlock() {
			while (index < nodeCount()) {
				Node node = node(index++);
				if (node instanceof ISOAPHeaderBlock) {
					return (ISOAPHeaderBlock) node;
				}
			}
			return null;
		}
	}

	public SOAPHeaderImpl(QName qname) {
		super(qname);
		if (qname == null) {
			throw new SOAPException("The qname should be not null.");
		}
		if (!qname.getName().equals(SOAPConstants.SOAP_HEADER_NAME)) {
			throw new SOAPException(
					"The SOAP Header name should be Header. But it is "
							+ qname.getName());
		}
	}

	public SOAPHeaderImpl(String name, Namespace namespace) {
		super(name, namespace);
		if (!name.equals(SOAPConstants.SOAP_HEADER_NAME)) {
			throw new SOAPException(
					"The SOAP Header name should be Header. But it is " + name);
		}
	}

	public SOAPHeaderImpl() {
		super(SOAPConstants.SOAP_HEADER_NAME);
	}

	public Iterator<ISOAPHeaderBlock> examineAllHeaderBlocks() {
		class DefaultChecker implements Checker {
			public boolean checkHeader(ISOAPHeaderBlock header) {
				return true;
			}
		}
		return new HeaderIterator(new DefaultChecker());
	}

	public Iterator<ISOAPHeaderBlock> examineHeaderBlocks(String role) {
		return new HeaderIterator(new RoleChecker(role));
	}

	public Iterator<ISOAPHeaderBlock> examineMustUnderstandHeaderBlocks(
			final String role) {
		return new HeaderIterator(new MURoleChecker(role));
	}

	public ArrayList<ISOAPHeaderBlock> getHeaderBlocksWithNSURI(String nsURI) {
		Iterator<Metaset> headers = elementIterator();
		if (headers.hasNext()) {
			ArrayList<ISOAPHeaderBlock> results = new ArrayList<ISOAPHeaderBlock>();
			while (headers.hasNext()) {
				Metaset header = headers.next();
				Namespace namespace = header.getNamespace();
				if (nsURI == null) {
					if (namespace == null) {
						results.add((ISOAPHeaderBlock) header);
					}
				} else {
					if (namespace != null) {
						if (nsURI.equals(namespace.getURI())) {
							results.add((ISOAPHeaderBlock) header);
						}
					}
				}
			}
			return results;
		} else {
			return null;
		}
	}

	public Iterator<ISOAPHeaderBlock> getHeadersToProcess(IRolePlayer rolePlayer) {
		return new HeaderIterator(new RolePlayerChecker(rolePlayer));
	}

	public Iterator<ISOAPHeaderBlock> getHeadersToProcess(
			IRolePlayer rolePlayer, String namespace) {
		return new HeaderIterator(new RolePlayerChecker(rolePlayer, namespace));
	}

	public Iterator<ISOAPHeaderBlock> extractAllHeaderBlocks() {
		class DefaultChecker implements Checker {
			public boolean checkHeader(ISOAPHeaderBlock header) {
				return true;
			}
		}
		return new HeaderIterator(new DefaultChecker(), true);
	}

	public Iterator<ISOAPHeaderBlock> extractHeaderBlocks(String role) {
		return new HeaderIterator(new RoleChecker(role), true);
	}

	public List<ISOAPHeaderBlock>getAllHeaderBlocks(){
		return headerBlocks();
	}
	
	private List<ISOAPHeaderBlock> headerBlocks(){
		BackedList<ISOAPHeaderBlock> answer = createResultList();
		for (Node node : contentList()) {
			appendElementLocal(node, answer);
		}
		return answer;
	}
	
	public  BackedList<ISOAPHeaderBlock> appendElementLocal(Node node, BackedList<ISOAPHeaderBlock> list) {
		ISOAPHeaderBlock header = (ISOAPHeaderBlock)NodeHelper.nodeAsElement(node);
		if (header != null) {
			list.addLocal(header);
		}

		return list;
	}
}
