package org.act.xservices.rt.xom.util;

import java.text.DateFormat;
import java.text.DecimalFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.TimeZone;

import org.act.xservices.rt.metaset.MetasetBuilder;
import org.act.xservices.rt.xom.Attribute;
import org.act.xservices.rt.xom.Metaset;
import org.act.xservices.rt.xom.Namespace;
import org.act.xservices.rt.xom.QName;

/**
 * 
 * Metaset helper class
 * 
 * @author douyl
 * 
 */
public class MetasetHelper {

	/**
	 * Singleton model
	 */
	private static MetasetHelper instance = null;

	private MetasetHelper() {
	}

	/**
	 * Get instance of MetasetHelper(Singleton mode)
	 * 
	 * @return An instance of MetasetHelper
	 */
	public static MetasetHelper getInstance() {
		if (instance == null) {
			synchronized (MetasetHelper.class) {
				if (instance == null) {
					instance = new MetasetHelper();
				}
			}
		}
		return instance;
	}

	/**
	 * Metaset has two kind of elements, a textNode & a elementNode TEXT_NODE
	 * represents the node who has only text ELEMENT_NODE represents the node
	 * who has children elements
	 */
	public static final int TEXT_NODE = 101;
	public static final int ELEMENT_NODE = 102;
	public static final int ERROR = -1;

	private static DateFormat _FORMATTER = new SimpleDateFormat(
			"yyyy-MM-dd'T'HH:mm:ss'Z'");

	/**
	 * Get type of metaset, there are only two type for metaset(TEXT_NODE &
	 * ELEMENT_NODE)
	 * 
	 * @param met
	 *            The metaset whose type will be returned
	 * @return The metaset type
	 * @see TEXT_NODE & ELEMENT_NODE
	 */
	public int getTypeOfMetaset(Metaset met) {
		if (met == null)
			return ERROR;

		if (met.elements().size() == 0)
			return TEXT_NODE;
		else
			return ELEMENT_NODE;
	}

	/**
	 * 
	 * @param e1
	 * @param e2
	 * 
	 * @return True, if:
	 *         <ul>
	 *         <li>The two Elements have the same QName.</li>
	 *         <li>They have the same child elements, <b>in the same order</b>;
	 *         this includes Elements and Texts.</li>
	 *         <li>They have the same attributes, excluding XML namespace and
	 *         prefix definitions.</li>
	 *         </ul>
	 *         This method short-circuits to "true" if the two references are
	 *         the same (e1 == e2). It short-circuits to "false" if one
	 *         reference is null and the other isn't.
	 * 
	 */
	public boolean isEqual(Metaset e1, Metaset e2) {
		//
		// avoid semantic comparisons if we can
		//
		if (e1 == e2)
			return true;

		//
		// they're not the same reference, so if either equals null,
		// we can quit
		//
		if (e1 == null || e2 == null)
			return false;

		//
		// make sure root elements are the same, including tag name
		//
		QName qname1 = e1.getQName();
		QName qname2 = e2.getQName();

		if (!qname1.equals(qname2))
			return false;

		return haveMatchingChildren(e1, e2) && haveMatchingAttributes(e1, e2);
	}

	/**
	 * 
	 * Recursively compares the children in the Element sub-trees to see if they
	 * are of equal name, value, and <b>order</b>. Only Element and Text nodes
	 * are compared.
	 * 
	 * @param e1
	 * @param e2
	 * 
	 * @return True if the two Elements have the same children, in the same
	 *         order. The comparison only targets Element and Text children, not
	 *         Attr or other DOM Nodes.
	 * 
	 * @see #haveMatchingAttributes(Element, Element)
	 * 
	 */
	private boolean haveMatchingChildren(Metaset e1, Metaset e2) {
		//
		// compare all the child nodes... (w/o attributes). nodes must be
		// identical and IN THE SAME ORDER - if this is so, the node types
		// should match
		//
		List<Metaset> children1 = e1.elements();
		List<Metaset> children2 = e2.elements();

		int length1 = children1.size();
		int length2 = children2.size();

		if (length1 != length2)
			return false;

		for (int n = 0; n < length1; ++n) {
			Metaset next1 = children1.get(n);
			Metaset next2 = children2.get(n);

			int type1 = this.getTypeOfMetaset(next1);
			int type2 = this.getTypeOfMetaset(next2);

			//
			// nodes aren't the same type - WRONG ORDER!
			//
			if (type1 != type2)
				return false;

			//
			// only concerned with Text and Element nodes - Comments and
			// CDATA are ignored
			//

			else if (type1 == TEXT_NODE) {
				String text1 = next1.getTextTrim();
				String text2 = next2.getTextTrim();

				if (!text1.equals(text2))
					return false;
			}

			//
			// recurse down the tree to do more comparisons
			//
			else if (type1 == ELEMENT_NODE
					&& !isEqual((Metaset) next1, (Metaset) next2))
				return false;
		}

		return true;
	}

	/**
	 * Used by isEquals(e1, e2)
	 * 
	 * @param e1
	 * @param e2
	 * 
	 * @return True if the two Elements have the same attributes and attribute
	 *         values. The comparison <b>excludes</b> XML namespace
	 *         declarations, since the prefixes used in an XML fragment are not
	 *         important.
	 * 
	 */
	private boolean haveMatchingAttributes(Metaset e1, Metaset e2) {
		//
		// the order of the attributes is not important, just the values.
		// we ignore namespace URI attributes because the prefixes used
		// in a fragment are also unimportant
		//
		List<Attribute> attr1 = e1.attributes();
		List<Attribute> attr2 = e2.attributes();

		int length1 = attr1.size();
		int length2 = attr2.size();

		if (length1 != length2)
			return false;

		for (int n = 0; n < length1; ++n) {
			Attribute attr = attr1.get(n);
			String name = attr.getName();
			String value = attr.getValue();

			Attribute match = getNamedItem(attr2, name);
			if (match == null)
				return false;

			String matchValue = match.getValue();

			if (matchValue == null || !matchValue.equals(value))
				return false;
		}

		return true;
	}

	/**
	 * Get item from attrList whose name equals param name
	 * 
	 * @param attrList
	 * @param name
	 * @return
	 */
	private Attribute getNamedItem(List<Attribute> attrList, String name) {
		Iterator<Attribute> iter = attrList.iterator();
		while (iter.hasNext()) {
			Attribute attr = iter.next();
			if (name.equalsIgnoreCase(attr.getName()))
				return attr;
		}
		return null;
	}

	/**
	 * Convert the qname to string
	 * 
	 * @param qname
	 * @return
	 */
	public String convertQNameToString(QName qname) {
		String prefix = qname.getNamespacePrefix();
		String name = qname.getName();

		if (prefix == null || prefix.length() == 0)
			return name;

		return prefix + ':' + name;
	}

	/**
	 * 
	 * Get direct child element from parentMetaset by childElemQName, and then
	 * get inner text from the child element (The inner text is a QName in text
	 * form)
	 * 
	 * @param xml
	 *            The root Element to perform the search on.
	 * 
	 * @param childQName
	 *            The name of the child element whose text is a QName.
	 * 
	 * @return A fully-resolved QName, or null if there was no text.
	 * 
	 */
	public QName getQNameFromChildElemText(Metaset parentMetaset,
			QName childElemQName) {
		//
		// we can't just get the text - we need to have the actual
		// element so we can resolve the QName's namespace
		//
		Metaset child = parentMetaset.element(childElemQName);

		if (child == null)
			return null;

		String qnameText = child.getTextTrim();
		if (qnameText == null || qnameText.length() == 0)
			return null;

		int colonIndex = qnameText.indexOf(':');
		String prefix = null;
		if (colonIndex >= 0)
			prefix = qnameText.substring(0, colonIndex);

		return convertStringToQName(qnameText, child);
	}

	/**
	 * Convert string in textNode to QName The textNode should be a
	 * textNode,which looks like <wsa:ParentItem xmlns:wsa="..."
	 * xmlns:tns="..."> tns:ChildItem </wsa:ParentItem>
	 * 
	 * @param textNode
	 * @return QName
	 */
	public QName convertStringToQName(String qname, Metaset textNode) {
		int colonIndex = qname.indexOf(':');

		String prefix = null;

		if (colonIndex >= 0)
			prefix = qname.substring(0, colonIndex);

		String localName = qname.substring(colonIndex + 1);

		String uri = null;
		Namespace nsQNamespace = null;

		//
		// if possible, try to resolve the prefix to a namespace URI
		//
		if (textNode != null)
			nsQNamespace = textNode.getNamespaceForPrefix(prefix);

		if (nsQNamespace == null)
			return null;

		//
		// prefix is not required, but it CANNOT be null
		//
		if (prefix != null && prefix.length() > 0)
			return new QName(nsQNamespace.getURI(), localName, prefix);

		return new QName(uri, localName);
	}

	/**
	 * Convert string in textNode to QName The textNode should be a
	 * textNode,which looks like <wsa:ParentItem xmlns:wsa="..."
	 * xmlns:tns="..."> tns:ChildItem </wsa:ParentItem>
	 * 
	 * @param textNode
	 * @return QName
	 */
	public QName convertStringToQName(Metaset textNode) {
		if (textNode == null || getTypeOfMetaset(textNode) == this.ELEMENT_NODE) {
			throw new RuntimeException("Convert exception occures");
		}

		String qnameText = textNode.getTextTrim();
		return convertStringToQName(qnameText, textNode);
	}

	/**
	 * 
	 * If the given context node has no child Elements with the given name, one
	 * is created; otherwise, the first direct child with the name is used.
	 * Either way, the match is assigned the given value, using the rules
	 * described in createElement(Document, QName, Object).
	 * 
	 * @see #createElement(Document, QName, Object)
	 * 
	 */
	public void setElement(Metaset context, QName qname, Object value) {
		if (context == null || qname == null || value == null)
			return;

		boolean embedChildren = value instanceof Metaset;
		Metaset met = context.element(qname);
		if (met == null) {
			met = MetasetBuilder.getInstance().createElement(qname);
			context.add(met);
		}

		if (embedChildren) {
			met.add((Metaset) value);
		} else if (value instanceof Date) {
			met.setText(getLocalTimeString((Date) value));
		} else {
			met.setText(value.toString());
		}
	}

	/**
	 * 
	 * Returns the text version of the given date, in standard XSD format. It
	 * will include a +/- suffix that adds or subtracts hours from the UTC time
	 * in order to express local time.
	 * 
	 * @param date
	 *            The date to serialize into text.
	 * 
	 * @return A string that describes the given date, in XSD format, with a +/-
	 *         suffix to denote the local time.
	 * 
	 */
	public String getLocalTimeString(Date date) {
		//
		// if we leave this date w/o a +N:00 postfix, calling code
		// will assume it is UTC
		//
		String dateWithoutTimeZone = _FORMATTER.format(date);

		//
		// figure out whether we're before (-) or after (+) UTC
		//
		TimeZone timeZone = _FORMATTER.getTimeZone();
		long offset = timeZone.getOffset(date.getTime());

		char sign = offset < 0 ? '-' : '+';
		offset = Math.abs(offset);

		//
		// figure out what comes after the +/-
		// 
		long hours = offset / 3600000;
		long minutes = (offset % 3600000) / 60000;

		//
		// add the +/- suffix and call it a day
		//        
		int length = dateWithoutTimeZone.length() + 6;
		StringBuffer buffer = new StringBuffer(length);

		DecimalFormat twoDigits = new DecimalFormat("00");

		buffer.append(dateWithoutTimeZone);
		buffer.append(sign);
		buffer.append(twoDigits.format(hours));
		buffer.append(':');
		buffer.append(twoDigits.format(minutes));

		return buffer.toString();
	}

	/**
	 * Convert dateTimeString to Date
	 * 
	 * @param dateTimeString
	 * @return Date
	 * @throws ParseException
	 */
	public Date getLocalTime(String dateTimeString) throws ParseException {
		return _FORMATTER.parse(dateTimeString);
	}

	public static void main(String[] args) throws ParseException {
		Date today = new Date();
		String dateString = MetasetHelper.getInstance().getLocalTimeString(today);
		System.out.println("dateString: " + dateString);
		
		Date thisDay = MetasetHelper.getInstance().getLocalTime(dateString);
		System.out.println("thisDay..........");
		System.out.println(thisDay.toLocaleString());
	}
}
