/**
 * ArithmeticFunctions.java
 */
package org.act.xservices.rt.xsecurity.xacml.pdp.function.impl;

import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.math.MathContext;
import java.math.RoundingMode;

import org.act.xservices.rt.xsecurity.xacml.context.IRequest;
import org.act.xservices.rt.xsecurity.xacml.exception.DataTypeMappingException;
import org.act.xservices.rt.xsecurity.xacml.exception.IndeterminateException;
import org.act.xservices.rt.xsecurity.xacml.pdp.function.XacmlFunctionAnnotation;
import org.act.xservices.rt.xsecurity.xacml.pdp.function.XacmlFunctionProviderAnnotation;
import org.act.xservices.rt.xsecurity.xacml.pdp.function.impl.ArgumentCheckUtil;
import org.act.xservices.rt.xsecurity.xacml.policy.target.AttributeValue;
import org.act.xservices.rt.xsecurity.xacml.util.Constants;

/**
 * ʵOASIS XACML淶2.0ĸ㺯ֱȽϺ
 * 	Appendix A.3.2 Arithmetic functions
 * 	Appendix A.3.6 Numeric comparison functions
 *  Appendix A.3.4 Numeric data-type conversion functions
 * 
 * 
 */
@XacmlFunctionProviderAnnotation
public class ArithmeticFunctions {
	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:integer-add",
			"urn:oasis:names:tc:xacml:1.0:function:double-add" })
	public static AttributeValue numberAdd(IRequest request,
			Object[] params) throws IndeterminateException {
		// OASIS XACML淶2.0integer-adddouble-addĲ>=2ﲻ
		ArgumentCheckUtil.checkArguments(params, -1);
		if (params.length < 2) {
			throw new IndeterminateException(Constants.STATUS_PROCESSINGERROR,
					"Invalid argument number. Expected 2 or more parameters, but got "
							+ params.length, null);
		}
		AttributeValue num = (AttributeValue) params[0];
		ArgumentCheckUtil.checkArgumentNumberType(num);

		try {
			Object realNum = num.getValue();
			Class<?> clazz = realNum.getClass();
			Method add = clazz.getMethod("add", clazz);
			for (int i = 1; i < params.length; i++) {
				AttributeValue number = (AttributeValue) params[i];
				ArgumentCheckUtil.checkArgumentNumberType(number);
				realNum = add.invoke(realNum, number.getValue());
			}
			return AttributeValue.newInstance(num.getDataType(), realNum);
		} catch (Exception e) {
			throw new IndeterminateException(Constants.STATUS_PROCESSINGERROR,
					"Error occurs in Number add.", e);
		}
	}

	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:integer-subtract",
			"urn:oasis:names:tc:xacml:1.0:function:double-subtract" })
	public static AttributeValue numberSubtract(IRequest request,
			Object[] params) throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 2);
		AttributeValue num1 = (AttributeValue) params[0];
		AttributeValue num2 = (AttributeValue) params[1];
		ArgumentCheckUtil.checkArgumentNumberType(num1);
		ArgumentCheckUtil.checkArgumentNumberType(num2);

		try {
			Object realNum = num1.getValue();
			Class<?> clazz = realNum.getClass();
			Method subtract = clazz.getMethod("subtract", clazz);
			return AttributeValue.newInstance(num1.getDataType(), subtract
					.invoke(realNum, num2.getValue()));
		} catch (Exception e) {
			throw new IndeterminateException(Constants.STATUS_PROCESSINGERROR,
					"Error occurs in Number subtract.", e);
		}
	}

	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:integer-multiply",
			"urn:oasis:names:tc:xacml:1.0:function:double-multiply" })
	public static AttributeValue numberMultiply(IRequest request,
			Object[] params) throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 2);
		AttributeValue num1 = (AttributeValue) params[0];
		AttributeValue num2 = (AttributeValue) params[1];
		ArgumentCheckUtil.checkArgumentNumberType(num1);
		ArgumentCheckUtil.checkArgumentNumberType(num2);

		try {
			Object realNum = num1.getValue();
			Class<?> clazz = realNum.getClass();
			Method multiply = clazz.getMethod("multiply", clazz);
			return AttributeValue.newInstance(num1.getDataType(), multiply
					.invoke(realNum, num2.getValue()));
		} catch (Exception e) {
			throw new IndeterminateException(Constants.STATUS_PROCESSINGERROR,
					"Error occurs in Number multiply.", e);
		}
	}

	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:integer-divide",
			"urn:oasis:names:tc:xacml:1.0:function:double-divide" })
	public static AttributeValue numberDivide(IRequest request,
			Object[] params) throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 2);
		AttributeValue num = (AttributeValue) params[0];
		AttributeValue divisor = (AttributeValue) params[1];
		ArgumentCheckUtil.checkArgumentNumberType(num);
		ArgumentCheckUtil.checkArgumentNumberType(divisor);

		if (((Number) divisor.getValue()).intValue() == 0) {
			throw new IndeterminateException(Constants.STATUS_PROCESSINGERROR,
					"Error occurs in Number add, Divided by zero.", null);
		}

		try {
			Object realNum = num.getValue();
			Class<?> clazz = realNum.getClass();
			Method divide = clazz.getMethod("divide", clazz);
			return AttributeValue.newInstance(num.getDataType(), divide.invoke(
					realNum, divisor.getValue()));
		} catch (Exception e) {
			throw new IndeterminateException(Constants.STATUS_PROCESSINGERROR,
					"Error occurs in Number add.", e);
		}
	}

	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:integer-abs",
			"urn:oasis:names:tc:xacml:1.0:function:double-abs" })
	public static AttributeValue numberAbs(IRequest request,
			Object[] params) throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 1);
		AttributeValue num = (AttributeValue) params[0];
		ArgumentCheckUtil.checkArgumentNumberType(num);

		try {
			Object realNum = num.getValue();
			Class<?> clazz = realNum.getClass();
			Method abs = clazz.getMethod("abs");
			return AttributeValue.newInstance(num.getDataType(), abs
					.invoke(realNum));
		} catch (Exception e) {
			throw new IndeterminateException(Constants.STATUS_PROCESSINGERROR,
					"Error occurs in Number abs.", e);
		}
	}

	@XacmlFunctionAnnotation(funcURIs = { "urn:oasis:names:tc:xacml:1.0:function:integer-mod" })
	public static AttributeValue integerMod(IRequest request,
			Object[] params) throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 2);
		AttributeValue num = (AttributeValue) params[0];
		AttributeValue divisor = (AttributeValue) params[1];
		ArgumentCheckUtil.checkArgumentType(num, Constants.TYPE_INTEGER);
		ArgumentCheckUtil.checkArgumentType(divisor, Constants.TYPE_INTEGER);

		if (((Number) divisor.getValue()).intValue() == 0) {
			throw new IndeterminateException(Constants.STATUS_PROCESSINGERROR,
					"Error occurs in Integer mod, Divided by zero.", null);
		}

		try {
			BigInteger i = (BigInteger) num.getValue();
			BigInteger d = (BigInteger) divisor.getValue();
			return AttributeValue.newInstance(num.getDataType(), i.mod(d));
		} catch (Exception e) {
			throw new IndeterminateException(Constants.STATUS_PROCESSINGERROR,
					"EError occurs in Integer mod.", e);
		}
	}

	@XacmlFunctionAnnotation(funcURIs = { "urn:oasis:names:tc:xacml:1.0:function:round" })
	public static AttributeValue round(IRequest request, Object[] params)
			throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 1);
		AttributeValue num = (AttributeValue) params[0];
		ArgumentCheckUtil.checkArgumentType(num, Constants.TYPE_DOUBLE);

		try {
			BigDecimal d = (BigDecimal) num.getValue();
			return AttributeValue.newInstance(num.getDataType(), d
					.round(new MathContext(0, RoundingMode.HALF_UP)));
		} catch (Exception e) {
			throw new IndeterminateException(Constants.STATUS_PROCESSINGERROR,
					"Error occurs in Double round.", e);
		}
	}

	@XacmlFunctionAnnotation(funcURIs = { "urn:oasis:names:tc:xacml:1.0:function:floor" })
	public static AttributeValue floor(IRequest request, Object[] params)
			throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 1);
		AttributeValue num = (AttributeValue) params[0];
		ArgumentCheckUtil.checkArgumentType(num, Constants.TYPE_DOUBLE);

		try {
			BigDecimal d = (BigDecimal) num.getValue();
			return AttributeValue.newInstance(num.getDataType(), d
					.round(new MathContext(0, RoundingMode.FLOOR)));
		} catch (Exception e) {
			throw new IndeterminateException(Constants.STATUS_PROCESSINGERROR,
					"Error occurs in Double floor.", e);
		}
	}
	
	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:integer-greater-than",
			"urn:oasis:names:tc:xacml:1.0:function:double-greater-than" })
	public static AttributeValue numberGreaterThan(IRequest request,
			Object[] params) throws IndeterminateException {
		return compareNumber(params) > 0 ? AttributeValue.TRUE
				: AttributeValue.FALSE;
	}

	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:integer-greater-than-or-equal",
			"urn:oasis:names:tc:xacml:1.0:function:double-greater-than-or-equal" })
	public static AttributeValue numberGreaterThanOrEqual(
			IRequest request, Object[] params)
			throws IndeterminateException {
		return compareNumber(params) >= 0 ? AttributeValue.TRUE
				: AttributeValue.FALSE;
	}

	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:integer-less-than",
			"urn:oasis:names:tc:xacml:1.0:function:double-less-than" })
	public static AttributeValue numberLessThan(IRequest request,
			Object[] params) throws IndeterminateException {
		return compareNumber(params) < 0 ? AttributeValue.TRUE
				: AttributeValue.FALSE;
	}

	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:integer-less-than-or-equal",
			"urn:oasis:names:tc:xacml:1.0:function:double-less-than-or-equal" })
	public static AttributeValue numberLessThanOrEqual(IRequest request,
			Object[] params) throws IndeterminateException {
		return compareNumber(params) <= 0 ? AttributeValue.TRUE
				: AttributeValue.FALSE;
	}

	@XacmlFunctionAnnotation(funcURIs = { "urn:oasis:names:tc:xacml:1.0:function:double-to-integer" })
	public static AttributeValue double2Integer(IRequest request,
			Object[] params) throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 1);
		AttributeValue d = (AttributeValue) params[0];
		ArgumentCheckUtil.checkArgumentType(d, Constants.TYPE_DOUBLE);

		try {
			BigDecimal num = (BigDecimal) d.getValue();
			return AttributeValue.newInstance(Constants.TYPE_INTEGER, num
					.toBigInteger());
		} catch (Exception e) {
			throw new IndeterminateException(Constants.STATUS_PROCESSINGERROR,
					"Error occurs in double-to-integer.", e);
		}
	}

	@XacmlFunctionAnnotation(funcURIs = { "urn:oasis:names:tc:xacml:1.0:function:integer-to-double" })
	public static AttributeValue integer2Double(IRequest request,
			Object[] params) throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 1);
		AttributeValue i = (AttributeValue) params[0];
		ArgumentCheckUtil.checkArgumentType(i, Constants.TYPE_INTEGER);

		try {
			BigInteger num = (BigInteger) i.getValue();
			return AttributeValue.newInstance(Constants.TYPE_DOUBLE, num
					.toString());
		} catch (Exception e) {
			throw new IndeterminateException(Constants.STATUS_PROCESSINGERROR,
					"Error occurs in integer-to-double.", e);
		}
	}

	@SuppressWarnings("unchecked")
	private static int compareNumber(Object[] params)
			throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 2);
		AttributeValue num1 = (AttributeValue) params[0];
		AttributeValue num2 = (AttributeValue) params[1];
		ArgumentCheckUtil.checkArgumentNumberType(num1);
		ArgumentCheckUtil.checkArgumentNumberType(num2);

		Object real1 = num1.getValue();
		Object real2 = num2.getValue();
		if (real1 instanceof Comparable) {
			return ((Comparable) real1).compareTo(real2);
		}
		throw new IndeterminateException(Constants.STATUS_PROCESSINGERROR,
				"Argument type is not Comparable", null);
	}
	
	public static void main(String[] args) throws DataTypeMappingException, IndeterminateException{
		AttributeValue a1=AttributeValue.newInstance(Constants.TYPE_DOUBLE, "3.1");
		AttributeValue a2=AttributeValue.newInstance(Constants.TYPE_DOUBLE, "3.0");
		AttributeValue a3=AttributeValue.newInstance(Constants.TYPE_DOUBLE, "3.1");
		AttributeValue a4=AttributeValue.newInstance(Constants.TYPE_DOUBLE, "3.1");
		AttributeValue r=numberAdd(null,new Object[]{a1,a2,a3,a4});
		System.out.println(r.getValue());
		AttributeValue r1=numberSubtract(null,new Object[]{a1,a2});
		System.out.println(r1.getValue());
	}
}
