/**
 * BaseFunctions.java
 */
package org.act.xservices.rt.xsecurity.xacml.pdp.function.impl;

import java.math.BigInteger;
import java.net.URI;
import java.util.HashSet;
import java.util.Set;

import org.act.xservices.rt.xsecurity.xacml.context.IRequest;
import org.act.xservices.rt.xsecurity.xacml.exception.DataTypeMappingException;
import org.act.xservices.rt.xsecurity.xacml.exception.IndeterminateException;
import org.act.xservices.rt.xsecurity.xacml.pdp.function.XacmlFunctionAnnotation;
import org.act.xservices.rt.xsecurity.xacml.pdp.function.XacmlFunctionProviderAnnotation;
import org.act.xservices.rt.xsecurity.xacml.policy.target.AttributeValue;
import org.act.xservices.rt.xsecurity.xacml.util.Constants;

/**
 * ʵOASIS XACML淶2.0һЩ:
 * 	Appendix A.3.1 Equality predicates
 *  Appendix A.3.10 Bag functions
 *  Appendix A.3.11 Set functions
 *  
 * 
 */
@XacmlFunctionProviderAnnotation
public class BaseFunctions {
	
	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:string-equal",
			"urn:oasis:names:tc:xacml:1.0:function:boolean-equal",
			"urn:oasis:names:tc:xacml:1.0:function:integer-equal",
			"urn:oasis:names:tc:xacml:1.0:function:date-equal",
			"urn:oasis:names:tc:xacml:1.0:function:time-equal",
			"urn:oasis:names:tc:xacml:1.0:function:dateTime-equal",
			"urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-equal",
			"urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-equal",
			"urn:oasis:names:tc:xacml:1.0:function:anyURI-equal",
			"urn:oasis:names:tc:xacml:1.0:function:x500Name-equal",
			"urn:oasis:names:tc:xacml:1.0:function:rfc822Name-equal",
			"urn:oasis:names:tc:xacml:1.0:function:hexBinary-equal",
			"urn:oasis:names:tc:xacml:1.0:function:base64Binary-equal" })
	public static AttributeValue equals(IRequest request, Object[] params)
			throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 2);
		AttributeValue o1 = (AttributeValue) params[0];
		AttributeValue o2 = (AttributeValue) params[1];
		if (o1 != null && o2 != null) {
			if (o1 == o2) {
				return AttributeValue.TRUE;
			}
			return o1.equals(o2) ? AttributeValue.TRUE : AttributeValue.FALSE;
		}
		return AttributeValue.FALSE;
	}

	
	/**
	 * @param request
	 * @param params
	 * @return
	 * @throws IndeterminateException
	 */
	@XacmlFunctionAnnotation(funcURIs = {
        "urn:oasis:names:tc:xacml:1.0:function:double-equal"
    })
    public static AttributeValue doubleEquals(IRequest request, Object[] params) throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 2);
        AttributeValue o1 = (AttributeValue)params[0];
        AttributeValue o2 = (AttributeValue)params[1];
        ArgumentCheckUtil.checkArgumentType(o1, Constants.TYPE_DOUBLE);
        ArgumentCheckUtil.checkArgumentType(o2, Constants.TYPE_DOUBLE);
        if (o1 != null && o2 != null) {
            if (o1 == o2) {
                return AttributeValue.TRUE;
            }
            Comparable double1 = (Comparable)o1.getValue();
            Object double2 = o2.getValue();
            return double1.compareTo(double2) == 0 ? AttributeValue.TRUE : AttributeValue.FALSE;
        }
        return AttributeValue.FALSE;
    }
	
	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:string-one-and-only",
			"urn:oasis:names:tc:xacml:1.0:function:boolean-one-and-only",
			"urn:oasis:names:tc:xacml:1.0:function:integer-one-and-only",
			"urn:oasis:names:tc:xacml:1.0:function:double-one-and-only",
			"urn:oasis:names:tc:xacml:1.0:function:time-one-and-only",
			"urn:oasis:names:tc:xacml:1.0:function:date-one-and-only",
			"urn:oasis:names:tc:xacml:1.0:function:dateTime-one-and-only",
			"urn:oasis:names:tc:xacml:1.0:function:anyURI-one-and-only",
			"urn:oasis:names:tc:xacml:1.0:function:hexBinary-one-and-only",
			"urn:oasis:names:tc:xacml:1.0:function:base64Binary-one-and-only",
			"urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-one-and-only",
			"urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-one-and-only",
			"urn:oasis:names:tc:xacml:1.0:function:x500Name-one-and-only",
			"urn:oasis:names:tc:xacml:1.0:function:rfc822Name-one-and-only" })
	public static AttributeValue bagOneAndOnly(IRequest request,
			Object[] params) throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 1);
		AttributeValue[] bag = (AttributeValue[]) params[0];
		if (bag != null && bag.length == 1) {
			return bag[0];
		}
		throw new IndeterminateException(
				Constants.STATUS_PROCESSINGERROR,
				"In the type-one-and-only function: More than one AttributeValue in the bag!",
				null);
	}

	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:string-bag",
			"urn:oasis:names:tc:xacml:1.0:function:boolean-bag",
			"urn:oasis:names:tc:xacml:1.0:function:integer-bag",
			"urn:oasis:names:tc:xacml:1.0:function:double-bag",
			"urn:oasis:names:tc:xacml:1.0:function:time-bag",
			"urn:oasis:names:tc:xacml:1.0:function:date-bag",
			"urn:oasis:names:tc:xacml:1.0:function:dateTime-bag",
			"urn:oasis:names:tc:xacml:1.0:function:anyURI-bag",
			"urn:oasis:names:tc:xacml:1.0:function:hexBinary-bag",
			"urn:oasis:names:tc:xacml:1.0:function:base64Binary-bag",
			"urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-bag",
			"urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-bag",
			"urn:oasis:names:tc:xacml:1.0:function:x500Name-bag",
			"urn:oasis:names:tc:xacml:1.0:function:rfc822Name-bag" })
	public static AttributeValue[] bagBag(IRequest request,
			Object[] params) throws IndeterminateException {
		for (Object param : params) {
			if (param != null && !(param instanceof AttributeValue)) {
				throw new IndeterminateException(
						Constants.STATUS_PROCESSINGERROR,
						"In the type-bag function: Some argument is null or not of AttributeValue type!",
						null);
			}
		}
		AttributeValue[] result = new AttributeValue[params.length];
		System.arraycopy(params, 0, result, 0, params.length);
		return result;
	}

	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:string-bag-size",
			"urn:oasis:names:tc:xacml:1.0:function:boolean-bag-size",
			"urn:oasis:names:tc:xacml:1.0:function:integer-bag-size",
			"urn:oasis:names:tc:xacml:1.0:function:double-bag-size",
			"urn:oasis:names:tc:xacml:1.0:function:time-bag-size",
			"urn:oasis:names:tc:xacml:1.0:function:date-bag-size",
			"urn:oasis:names:tc:xacml:1.0:function:dateTime-bag-size",
			"urn:oasis:names:tc:xacml:1.0:function:anyURI-bag-size",
			"urn:oasis:names:tc:xacml:1.0:function:hexBinary-bag-size",
			"urn:oasis:names:tc:xacml:1.0:function:base64Binary-bag-size",
			"urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-bag-size",
			"urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-bag-size",
			"urn:oasis:names:tc:xacml:1.0:function:x500Name-bag-size",
			"urn:oasis:names:tc:xacml:1.0:function:rfc822Name-bag-size" })
	public static AttributeValue bagSize(IRequest request,
			Object[] params) throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 1);
		AttributeValue[] bag = (AttributeValue[]) params[0];
		try {
			return AttributeValue.newInstance(Constants.TYPE_INTEGER,
					BigInteger.valueOf(bag.length));
		} catch (DataTypeMappingException e) {
			// TODO Auto-generated catch block
			throw new IndeterminateException(Constants.STATUS_PROCESSINGERROR,
					"In the type-bag-size function: " + e.getMessage(), e);
		}
	}

	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:string-is-in",
			"urn:oasis:names:tc:xacml:1.0:function:boolean-is-in",
			"urn:oasis:names:tc:xacml:1.0:function:integer-is-in",
			"urn:oasis:names:tc:xacml:1.0:function:double-is-in",
			"urn:oasis:names:tc:xacml:1.0:function:time-is-in",
			"urn:oasis:names:tc:xacml:1.0:function:date-is-in",
			"urn:oasis:names:tc:xacml:1.0:function:dateTime-is-in",
			"urn:oasis:names:tc:xacml:1.0:function:anyURI-is-in",
			"urn:oasis:names:tc:xacml:1.0:function:hexBinary-is-in",
			"urn:oasis:names:tc:xacml:1.0:function:base64Binary-is-in",
			"urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-is-in",
			"urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-is-in",
			"urn:oasis:names:tc:xacml:1.0:function:x500Name-is-in",
			"urn:oasis:names:tc:xacml:1.0:function:rfc822Name-is-in" })
	public static AttributeValue bagIsIn(IRequest request,
			Object[] params) throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 2);
		ArgumentCheckUtil.checkNull(params);
		AttributeValue o = (AttributeValue) params[0];
		AttributeValue[] bag = (AttributeValue[]) params[1];

		for (AttributeValue each : bag) {
			if (equals(request, new AttributeValue[] { o, each }) == AttributeValue.TRUE) {
				return AttributeValue.TRUE;
			}
		}
		return AttributeValue.FALSE;
	}

	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:string-intersection",
			"urn:oasis:names:tc:xacml:1.0:function:boolean-intersection",
			"urn:oasis:names:tc:xacml:1.0:function:integer-intersection",
			"urn:oasis:names:tc:xacml:1.0:function:double-intersection",
			"urn:oasis:names:tc:xacml:1.0:function:time-intersection",
			"urn:oasis:names:tc:xacml:1.0:function:date-intersection",
			"urn:oasis:names:tc:xacml:1.0:function:dateTime-intersection",
			"urn:oasis:names:tc:xacml:1.0:function:anyURI-intersection",
			"urn:oasis:names:tc:xacml:1.0:function:hexBinary-intersection",
			"urn:oasis:names:tc:xacml:1.0:function:base64Binary-intersection",
			"urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-intersection",
			"urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-intersection",
			"urn:oasis:names:tc:xacml:1.0:function:x500Name-intersection",
			"urn:oasis:names:tc:xacml:1.0:function:rfc822Name-intersection" })
	public static AttributeValue[] setIntersection(IRequest request,
			Object[] params) throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 2);
		AttributeValue[] bag1 = (AttributeValue[]) params[0];
		AttributeValue[] bag2 = (AttributeValue[]) params[0];

		Set<AttributeValue> result = new HashSet<AttributeValue>();
		for (AttributeValue each1 : bag1) {
			for (AttributeValue each2 : bag2) {
				if (equals(request, new AttributeValue[] { each1, each2 }) == AttributeValue.TRUE) {
					result.add(each1);
				}
			}
		}
		return result.toArray(bag1);
	}

	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:string-union",
			"urn:oasis:names:tc:xacml:1.0:function:boolean-union",
			"urn:oasis:names:tc:xacml:1.0:function:integer-union",
			"urn:oasis:names:tc:xacml:1.0:function:double-union",
			"urn:oasis:names:tc:xacml:1.0:function:time-union",
			"urn:oasis:names:tc:xacml:1.0:function:date-union",
			"urn:oasis:names:tc:xacml:1.0:function:dateTime-union",
			"urn:oasis:names:tc:xacml:1.0:function:anyURI-union",
			"urn:oasis:names:tc:xacml:1.0:function:hexBinary-union",
			"urn:oasis:names:tc:xacml:1.0:function:base64Binary-union",
			"urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-union",
			"urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-union",
			"urn:oasis:names:tc:xacml:1.0:function:x500Name-union",
			"urn:oasis:names:tc:xacml:1.0:function:rfc822Name-union" })
	public static AttributeValue[] setUnion(IRequest request,
			Object[] params) throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 2);
		AttributeValue[] bag1 = (AttributeValue[]) params[0];
		AttributeValue[] bag2 = (AttributeValue[]) params[0];

		Set<AttributeValue> result = new HashSet<AttributeValue>();
		for (AttributeValue each : bag1) {
			result.add(each);
		}
		for (AttributeValue each : bag2) {
			result.add(each);
		}
		return result.toArray(bag1);
	}

	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:string-subset",
			"urn:oasis:names:tc:xacml:1.0:function:boolean-subset",
			"urn:oasis:names:tc:xacml:1.0:function:integer-subset",
			"urn:oasis:names:tc:xacml:1.0:function:double-subset",
			"urn:oasis:names:tc:xacml:1.0:function:time-subset",
			"urn:oasis:names:tc:xacml:1.0:function:date-subset",
			"urn:oasis:names:tc:xacml:1.0:function:dateTime-subset",
			"urn:oasis:names:tc:xacml:1.0:function:anyURI-subset",
			"urn:oasis:names:tc:xacml:1.0:function:hexBinary-subset",
			"urn:oasis:names:tc:xacml:1.0:function:base64Binary-subset",
			"urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-subset",
			"urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-subset",
			"urn:oasis:names:tc:xacml:1.0:function:x500Name-subset",
			"urn:oasis:names:tc:xacml:1.0:function:rfc822Name-subset" })
	public static AttributeValue setSubset(IRequest request,
			Object[] params) throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 2);
		AttributeValue[] bag1 = (AttributeValue[]) params[0];
		AttributeValue[] bag2 = (AttributeValue[]) params[0];

		Set<AttributeValue> set1 = new HashSet<AttributeValue>();
		for (AttributeValue each : bag1) {
			set1.add(each);
		}
		Set<AttributeValue> set2 = new HashSet<AttributeValue>();
		for (AttributeValue each : bag2) {
			set2.add(each);
		}
		return set2.containsAll(set1) ? AttributeValue.TRUE
				: AttributeValue.FALSE;
	}

	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:string-set-equals",
			"urn:oasis:names:tc:xacml:1.0:function:boolean-set-equals",
			"urn:oasis:names:tc:xacml:1.0:function:integer-set-equals",
			"urn:oasis:names:tc:xacml:1.0:function:double-set-equals",
			"urn:oasis:names:tc:xacml:1.0:function:time-set-equals",
			"urn:oasis:names:tc:xacml:1.0:function:date-set-equals",
			"urn:oasis:names:tc:xacml:1.0:function:dateTime-set-equals",
			"urn:oasis:names:tc:xacml:1.0:function:anyURI-set-equals",
			"urn:oasis:names:tc:xacml:1.0:function:hexBinary-set-equals",
			"urn:oasis:names:tc:xacml:1.0:function:base64Binary-set-equals",
			"urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-set-equals",
			"urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-set-equals",
			"urn:oasis:names:tc:xacml:1.0:function:x500Name-set-equals",
			"urn:oasis:names:tc:xacml:1.0:function:rfc822Name-set-equals" })
	public static AttributeValue setEquals(IRequest request,
			Object[] params) throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 2);
		AttributeValue[] bag1 = (AttributeValue[]) params[0];
		AttributeValue[] bag2 = (AttributeValue[]) params[0];

		// FIXME - I didn't follow the specification here.
		Set<AttributeValue> set1 = new HashSet<AttributeValue>();
		for (AttributeValue each : bag1) {
			set1.add(each);
		}
		Set<AttributeValue> set2 = new HashSet<AttributeValue>();
		for (AttributeValue each : bag2) {
			set2.add(each);
		}
		return (set1.size() == set2.size() && set1.containsAll(set2)) ? AttributeValue.TRUE
				: AttributeValue.FALSE;
	}

	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:string-at-least-one-member-of",
			"urn:oasis:names:tc:xacml:1.0:function:boolean-at-least-one-member-of",
			"urn:oasis:names:tc:xacml:1.0:function:integer-at-least-one-member-of",
			"urn:oasis:names:tc:xacml:1.0:function:double-at-least-one-member-of",
			"urn:oasis:names:tc:xacml:1.0:function:time-at-least-one-member-of",
			"urn:oasis:names:tc:xacml:1.0:function:date-at-least-one-member-of",
			"urn:oasis:names:tc:xacml:1.0:function:dateTime-at-least-one-member-of",
			"urn:oasis:names:tc:xacml:1.0:function:anyURI-at-least-one-member-of",
			"urn:oasis:names:tc:xacml:1.0:function:hexBinary-at-least-one-member-of",
			"urn:oasis:names:tc:xacml:1.0:function:base64Binary-at-least-one-member-of",
			"urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-at-least-one-member-of",
			"urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-at-least-one-member-of",
			"urn:oasis:names:tc:xacml:1.0:function:x500Name-at-least-one-member-of",
			"urn:oasis:names:tc:xacml:1.0:function:rfc822Name-at-least-one-member-of" })
	public static AttributeValue setAtLeastOneMemberOf(IRequest request,
			Object[] params) throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 2);
		AttributeValue[] bag1 = (AttributeValue[]) params[0];
		AttributeValue[] bag2 = (AttributeValue[]) params[1];

		for (AttributeValue each : bag1) {
			if (bagIsIn(request, new Object[] { each, bag2 }) == AttributeValue.TRUE) {
				return AttributeValue.TRUE;
			}
		}
		return AttributeValue.FALSE;
	}
}
