/**
 * DateTimeFunctions.java
 */
package org.act.xservices.rt.xsecurity.xacml.pdp.function.impl;

import java.net.URI;

import javax.xml.datatype.Duration;
import javax.xml.datatype.XMLGregorianCalendar;

import org.act.xservices.rt.xsecurity.xacml.context.IRequest;
import org.act.xservices.rt.xsecurity.xacml.exception.DataTypeMappingException;
import org.act.xservices.rt.xsecurity.xacml.exception.IndeterminateException;
import org.act.xservices.rt.xsecurity.xacml.pdp.function.XacmlFunctionAnnotation;
import org.act.xservices.rt.xsecurity.xacml.pdp.function.XacmlFunctionProviderAnnotation;
import org.act.xservices.rt.xsecurity.xacml.policy.target.AttributeValue;
import org.act.xservices.rt.xsecurity.xacml.util.Constants;

/**
 * ʵOASIS XACML淶2.0ںʱصĺ
 *  Appendix A.3.7 Date and time arithmetic functions
 *  Appendix A.3.8 Non-numeric comparison functionsеںʱȽ
 *  
 * 
 */

@XacmlFunctionProviderAnnotation
public class DateTimeFunctions {

	private static int compareDate(Object[] params)
			throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 2);
		AttributeValue date1 = (AttributeValue) params[0];
		AttributeValue date2 = (AttributeValue) params[1];
		ArgumentCheckUtil.checkArgumentDateType(date1);
		ArgumentCheckUtil.checkArgumentDateType(date2);

		XMLGregorianCalendar actualDate1 = (XMLGregorianCalendar) date1
				.getValue();
		XMLGregorianCalendar actualDate2 = (XMLGregorianCalendar) date2
				.getValue();
		return actualDate1.compare(actualDate2);
	}

	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:dateTime-add-dayTimeDuration",
			"urn:oasis:names:tc:xacml:1.0:function:dateTime-add-yearMonthDuration",
			"urn:oasis:names:tc:xacml:1.0:function:date-add-yearMonthDuration" })
	public static AttributeValue dateAdd(IRequest request,
			Object[] params) throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 2);
		AttributeValue date = (AttributeValue) params[0];
		AttributeValue duration = (AttributeValue) params[1];
		ArgumentCheckUtil.checkArgumentDateType(date);
		ArgumentCheckUtil.checkArgumentDurationType(duration);

		try {
			XMLGregorianCalendar cloned = (XMLGregorianCalendar) ((XMLGregorianCalendar) date
					.getValue()).clone();
			Duration actualDuration = (Duration) duration.getValue();
			cloned.add(actualDuration);

			return AttributeValue.newInstance(date.getDataType(), cloned);
		} catch (DataTypeMappingException e) {
			// TODO Auto-generated catch block
			throw new IndeterminateException(Constants.STATUS_PROCESSINGERROR,
					"In the dateTime-add-xxxDuration function: "
							+ e.getMessage(), null);
		}
	}

	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:dateTime-subtract-dayTimeDuration",
			"urn:oasis:names:tc:xacml:1.0:function:dateTime-subtract-yearMonthDuration",
			"urn:oasis:names:tc:xacml:1.0:function:date-subtract-yearMonthDuration" })
	public static AttributeValue dateSubtract(IRequest request,
			Object[] params) throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 2);
		AttributeValue date = (AttributeValue) params[0];
		AttributeValue duration = (AttributeValue) params[1];
		ArgumentCheckUtil.checkArgumentDateType(date);
		ArgumentCheckUtil.checkArgumentDurationType(duration);

		try {
			XMLGregorianCalendar cloned = (XMLGregorianCalendar) ((XMLGregorianCalendar) date
					.getValue()).clone();
			Duration actualDuration = (Duration) duration.getValue();
			cloned.add(actualDuration.negate());

			return AttributeValue.newInstance(date.getDataType(), cloned);
		} catch (DataTypeMappingException e) {
			// TODO Auto-generated catch block
			throw new IndeterminateException(Constants.STATUS_PROCESSINGERROR,
					"In the dateTime-subtract-xxxDuration function: "
							+ e.getMessage(), null);
		}
	}

	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:date-greater-than",
			"urn:oasis:names:tc:xacml:1.0:function:time-greater-than",
			"urn:oasis:names:tc:xacml:1.0:function:dateTime-greater-than" })
	public static AttributeValue dateGreaterThan(IRequest request,
			Object[] params) throws IndeterminateException {
		return compareDate(params) > 0 ? AttributeValue.TRUE
				: AttributeValue.FALSE;
	}

	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:date-greater-than-or-equal",
			"urn:oasis:names:tc:xacml:1.0:function:time-greater-than-or-equal",
			"urn:oasis:names:tc:xacml:1.0:function:dateTime-greater-than-or-equal" })
	public static AttributeValue dateGreaterThanOrEqual(
			IRequest request, Object[] params)
			throws IndeterminateException {
		return compareDate(params) >= 0 ? AttributeValue.TRUE
				: AttributeValue.FALSE;
	}

	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:date-less-than",
			"urn:oasis:names:tc:xacml:1.0:function:time-less-than",
			"urn:oasis:names:tc:xacml:1.0:function:dateTime-less-than" })
	public static AttributeValue dateLessThan(IRequest request,
			Object[] params) throws IndeterminateException {
		return compareDate(params) < 0 ? AttributeValue.TRUE
				: AttributeValue.FALSE;
	}

	@XacmlFunctionAnnotation(funcURIs = {
			"urn:oasis:names:tc:xacml:1.0:function:date-less-than-or-equal",
			"urn:oasis:names:tc:xacml:1.0:function:time-less-than-or-equal",
			"urn:oasis:names:tc:xacml:1.0:function:dateTime-less-than-or-equal" })
	public static AttributeValue dateLessThanOrEqual(IRequest request,
			Object[] params) throws IndeterminateException {
		return compareDate(params) <= 0 ? AttributeValue.TRUE
				: AttributeValue.FALSE;
	}

	@XacmlFunctionAnnotation(funcURIs = { "urn:oasis:names:tc:xacml:1.0:function:time-in-range" })
	public static AttributeValue timeInRange(IRequest request,
			Object[] params) throws IndeterminateException {
		ArgumentCheckUtil.checkArguments(params, 3);
		AttributeValue t1 = (AttributeValue) params[0];
		AttributeValue t2 = (AttributeValue) params[1];
		AttributeValue t3 = (AttributeValue) params[2];
		ArgumentCheckUtil.checkArgumentType(t1, Constants.TYPE_TIME);
		ArgumentCheckUtil.checkArgumentType(t2, Constants.TYPE_TIME);
		ArgumentCheckUtil.checkArgumentType(t3, Constants.TYPE_TIME);

		XMLGregorianCalendar time1 = (XMLGregorianCalendar) t1.getValue();
		XMLGregorianCalendar time2 = (XMLGregorianCalendar) t2.getValue();
		XMLGregorianCalendar time3 = (XMLGregorianCalendar) t3.getValue();

		if (time2.compare(time3) > 0) {
			throw new IndeterminateException(Constants.STATUS_PROCESSINGERROR,
					"Invalid time range : [" + time2 + " - " + time3 + "]",
					null);
		}

		return time1.compare(time2) >= 0 && time1.compare(time3) <= 0 ? AttributeValue.TRUE
				: AttributeValue.FALSE;
	}

}
