/**
 * Policy.java
 */
package org.act.xservices.rt.xsecurity.xacml.policy;

import java.net.URI;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.act.xservices.rt.xom.Attribute;
import org.act.xservices.rt.xom.Metaset;
import org.act.xservices.rt.xsecurity.xacml.context.Decision;
import org.act.xservices.rt.xsecurity.xacml.context.IRequest;
import org.act.xservices.rt.xsecurity.xacml.context.Result;
import org.act.xservices.rt.xsecurity.xacml.context.Status;
import org.act.xservices.rt.xsecurity.xacml.context.StatusCode;
import org.act.xservices.rt.xsecurity.xacml.exception.DataTypeMappingException;
import org.act.xservices.rt.xsecurity.xacml.exception.IndeterminateException;
import org.act.xservices.rt.xsecurity.xacml.exception.PolicySyntaxErrorException;
import org.act.xservices.rt.xsecurity.xacml.exception.XacmlException;
import org.act.xservices.rt.xsecurity.xacml.pdp.function.FunctionRegistry;
import org.act.xservices.rt.xsecurity.xacml.pdp.function.XacmlFunction;
import org.act.xservices.rt.xsecurity.xacml.policy.rule.Rule;
import org.act.xservices.rt.xsecurity.xacml.policy.rule.VariableDefinition;
import org.act.xservices.rt.xsecurity.xacml.policy.target.Target;
import org.act.xservices.rt.xsecurity.xacml.util.Constants;
import org.act.xservices.rt.xsecurity.xacml.util.LoggerFactory;

/**
 * @author huoyg [huoyg@act.buaa.edu.cn]
 * @version 1.0
 * @date 2009-4-29
 */
public class Policy extends AXacmlPolicy {

	public static final String VARIABLE_DEF = "VariableDefinition";

	public static final String RULE_COMB_ID = "RuleCombiningAlgId";

	public static final String RULE = "Rule";
	
	private static Logger logger=LoggerFactory.getLogger(Policy.class);

	private Map<String, VariableDefinition> variableDefMap = null;

	private Rule[] rules = null;

	private URI ruleCombiningAlgId = null;

	public Policy() {
		// PolicyresolvedΪtrue;
		this.resolved = true;
		this.variableDefMap = new HashMap<String, VariableDefinition>();
	}

	public Policy(URI id, String description, XacmlPolicyDefaults defaults,
			Target target, Obligations obligations, Rule[] rules,
			URI ruleCombiningAlgId) {
		super(id, description, defaults, target, obligations);
		this.rules = rules;
		for (int i = 0; i < rules.length; i++) {
			this.rules[i].setParent(this);
		}
		this.ruleCombiningAlgId = ruleCombiningAlgId;
		// PolicyresolvedΪtrue;
		this.resolved = true;
		this.variableDefMap = new HashMap<String, VariableDefinition>();
	}

	public Policy(Metaset xmlElement) throws PolicySyntaxErrorException,
			DataTypeMappingException, IndeterminateException {
		super(xmlElement);

		Metaset[] ruleChildren = UtilMetaset.extractChildrenByName(xmlElement,
				RULE);
		this.rules = new Rule[ruleChildren.length];
		for (int i = 0; i < this.rules.length; i++) {
			this.rules[i] = new Rule(ruleChildren[i]);
			this.rules[i].setParent(this);
		}

		Attribute[] combIdAttrs = UtilMetaset.extractAttributesByName(
				xmlElement, RULE_COMB_ID);
		UtilMetaset.checkEmpty(combIdAttrs, RULE_COMB_ID);
		UtilMetaset.checkNumber(combIdAttrs, 1, RULE_COMB_ID);
		String idStr = combIdAttrs[0].getValue().trim();
		this.ruleCombiningAlgId = URI.create(idStr);

		this.resolved = true;

		Metaset[] vDefChildren = UtilMetaset.extractChildrenByName(xmlElement,
				VARIABLE_DEF);
		for (int i = 0; i < vDefChildren.length; i++) {
			VariableDefinition vDef = new VariableDefinition(vDefChildren[i]);
			vDef.setParent(this);
			this.variableDefMap.put(vDef.getVariableId(), vDef);
		}
	}

	// һ塣
	public void addVariableDefinition(VariableDefinition variableDef)
			throws PolicySyntaxErrorException {
		String vId = variableDef.getVariableId();
		if (this.variableDefMap.get(vId) != null) {
			throw new PolicySyntaxErrorException(Constants.STATUS_SYNTAXERROR,
					"Variable re-definition!", null);
		}
		variableDef.setParent(this);
		this.variableDefMap.put(vId, variableDef);
	}

	// ɱIDȡ塣
	public VariableDefinition getVariableDef(String variableId) {
		return this.variableDefMap.get(variableId);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.act.xacml.policy.AXacmlPolicy#evaluate(org.act.xacml.context.IRequestContext)
	 *      ̡ 1.
	 *      ԵtargetΪIndeterminateNotMatchӦΪIntermediateNotApplicable;
	 *      2. targetΪmatchtarget.match(request)true: a.
	 *      ԰RuleΪ0һRuleΪEffectInterminateԵɺϲ㷨
	 *      b. Rule0ȫΪNotApplicableԵΪNotApplicable
	 * 
	 * οOASIS XACML淶2.07.10(Policy evaluation).
	 * עڹ淶еĺϲ㷨2bԣҲɺϲ㷨ûԶϲ㷨ҲӦ2b
	 */
	public Result evaluate(IRequest request) throws DataTypeMappingException {
		// TODO Auto-generated method stub
		boolean isMatch;
		try {
			if (this.target == null) {
				isMatch = true;
			} else {
				isMatch = this.target.match(request);
			}
		} catch (IndeterminateException e) {
			// targetΪIndeterminatePolicyΪIndeterminate
			logger.throwing("Policy", "evaluate", e);
			Result result = new Result();
			result.setDecision(Decision.INDETERMINATE);
			result.setStatus(Status.newStatus(e));
			return result;
		} 
		// targetΪNotMatchPolicyΪNotApplicable.
		if (!isMatch) {
			logger.logp(Level.FINE, "Policy", "evaluate", "Target not match, so the policy evaluates to NotApplicable.");
			Result result = new Result();
			result.setDecision(Decision.NOTAPPLICABLE);
			result.setStatus(new Status(StatusCode.STATUS_OK, ""));
			return result;
		}
		// targetѾƥ
		try {
			XacmlFunction ruleCombineAlg = FunctionRegistry.getInstance()
					.lookup(this.ruleCombiningAlgId);
			Result result = (Result) ruleCombineAlg.invoke(request, this.rules);
			if(this.obligations!=null){
				result.addObligations(this.obligations
						.getObligationsByDecision(result.getDecision()));
			}
			return result;
		} catch (XacmlException e) {
			// TODO Auto-generated catch block
			logger.throwing("Policy", "evaluate", e);
			Result result = new Result();
			result.setDecision(Decision.INDETERMINATE);
			result.setStatus(Status.newStatus(e));
			return result;
		}
	}

	public URI getRuleCombiningAlgId() {
		return ruleCombiningAlgId;
	}

	public void setRuleCombiningAlgId(URI ruleCombiningAlgId) {
		this.ruleCombiningAlgId = ruleCombiningAlgId;
	}

	public Rule[] getRules() {
		return rules;
	}

	public void setRules(Rule[] rules) {
		this.rules = rules;
	}

	public Map<String, VariableDefinition> getVariableDefMap() {
		return variableDefMap;
	}

	public void setVariableDefMap(Map<String, VariableDefinition> variableDefMap) {
		this.variableDefMap = variableDefMap;
		for (VariableDefinition vDef : this.variableDefMap.values()) {
			vDef.setParent(this);
		}
	}
}
