/**
 * Rule.java
 */
package org.act.xservices.rt.xsecurity.xacml.policy.rule;

import java.net.URI;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.act.xservices.rt.xom.Attribute;
import org.act.xservices.rt.xom.Metaset;
import org.act.xservices.rt.xsecurity.xacml.AXacmlElement;
import org.act.xservices.rt.xsecurity.xacml.IEvaluateable;
import org.act.xservices.rt.xsecurity.xacml.context.Decision;
import org.act.xservices.rt.xsecurity.xacml.context.IRequest;
import org.act.xservices.rt.xsecurity.xacml.context.Result;
import org.act.xservices.rt.xsecurity.xacml.context.Status;
import org.act.xservices.rt.xsecurity.xacml.context.StatusCode;
import org.act.xservices.rt.xsecurity.xacml.exception.DataTypeMappingException;
import org.act.xservices.rt.xsecurity.xacml.exception.IndeterminateException;
import org.act.xservices.rt.xsecurity.xacml.exception.PolicySyntaxErrorException;
import org.act.xservices.rt.xsecurity.xacml.policy.UtilMetaset;
import org.act.xservices.rt.xsecurity.xacml.policy.target.Target;
import org.act.xservices.rt.xsecurity.xacml.util.Constants;
import org.act.xservices.rt.xsecurity.xacml.util.LoggerFactory;

/**
 * 
 */
public class Rule extends AXacmlElement implements IEvaluateable {

	public static final String RULE_ID = "RuleId";

	public static final String EFFECT = "Effect";

	public static final String DESCRIPTION = "Description";

	public static final String TARGET = "Target";

	public static final String CONDITION = "Condition";

	public static final String PERMIT = "Permit";

	public static final String DENY = "Deny";
	
	private static Logger logger=LoggerFactory.getLogger(Rule.class);

	private URI ruleId = null;

	private Effect effect = null;

	private String description = null;

	private Target target = null;

	private Condition condition = null;

	public Rule() {
	}

	public Rule(URI ruleId, Effect effect, String description, Target target,
			Condition condition) {
		this.ruleId = ruleId;
		this.effect = effect;
		this.description = description;
		this.target = target;
		((AXacmlElement) this.target).setParent(this);
		this.condition = condition;
		((AXacmlElement) this.condition).setParent(this);
	}

	public Rule(Metaset xmlElement) throws PolicySyntaxErrorException,
			DataTypeMappingException, IndeterminateException {
		super(xmlElement);

		Attribute[] ruleIdAttrs = UtilMetaset.extractAttributesByName(
				xmlElement, RULE_ID);
		UtilMetaset.checkEmpty(ruleIdAttrs, RULE_ID);
		UtilMetaset.checkNumber(ruleIdAttrs, 1, RULE_ID);
		String ruleIdStr = ruleIdAttrs[0].getValue().trim();
		this.ruleId = URI.create(ruleIdStr);

		Attribute[] effectAtrrs = UtilMetaset.extractAttributesByName(
				xmlElement, EFFECT);
		UtilMetaset.checkEmpty(effectAtrrs, EFFECT);
		UtilMetaset.checkNumber(effectAtrrs, 1, EFFECT);
		String effectStr = effectAtrrs[0].getValue().trim();
		if (effectStr.equalsIgnoreCase(PERMIT)) {
			this.effect = Effect.PERMIT;
		} else if (effectStr.equalsIgnoreCase(DENY)) {
			this.effect = Effect.DENY;
		} else {
			throw new PolicySyntaxErrorException(Constants.STATUS_SYNTAXERROR,
					"Invalid effect value!", null);
		}

		Metaset[] descChildren = UtilMetaset.extractChildrenByName(xmlElement,
				DESCRIPTION);
		if (descChildren != null && descChildren.length != 0) {
			UtilMetaset.checkNumber(descChildren, 1, DESCRIPTION);
			this.description = descChildren[0].getText();
		}

		Metaset[] targetChildren = UtilMetaset.extractChildrenByName(
				xmlElement, TARGET);
		if (targetChildren != null && targetChildren.length != 0) {
			UtilMetaset.checkNumber(targetChildren, 1, TARGET);
			this.target = new Target(targetChildren[0]);
			((AXacmlElement) this.target).setParent(this);
		}

		Metaset[] condChildren = UtilMetaset.extractChildrenByName(xmlElement,
				CONDITION);
		if (condChildren != null && condChildren.length != 0) {
			UtilMetaset.checkNumber(condChildren, 1, CONDITION);
			this.condition = new Condition(condChildren[0]);
			((AXacmlElement) this.condition).setParent(this);
		}
	}

	public Condition getCondition() {
		return condition;
	}

	public void setCondition(Condition condition) {
		this.condition = condition;
		((AXacmlElement) this.condition).setParent(this);
	}

	public String getDescription() {
		return description;
	}

	public void setDescription(String description) {
		this.description = description;
	}

	public Effect getEffect() {
		return effect;
	}

	public void setEffect(Effect effect) {
		this.effect = effect;
	}

	public URI getRuleId() {
		return ruleId;
	}

	public void setRuleId(URI ruleId) {
		this.ruleId = ruleId;
	}

	public Target getTarget() {
		return target;
	}

	public void setTarget(Target target) {
		this.target = target;
		((AXacmlElement) this.target).setParent(this);
	}

	public Result evaluate(IRequest request) throws DataTypeMappingException {
		// TODO Auto-generated method stub
		Result result = new Result();
		boolean isMatch;
		try {
			if (this.target == null) {
				isMatch = true;
			} else {
				isMatch = this.target.match(request);
			}
		} catch (IndeterminateException ie) {
			// targetΪIndeterminateRuleΪIndeterminate
			logger.throwing("Rule", "evalue", ie);
			result.setDecision(Decision.INDETERMINATE);
			result.setStatus(Status.newStatus(ie));
			return result;
		} 
		// targetΪNotMatchRuleΪNotApplicable.
		if (!isMatch) {
			logger.logp(Level.FINE, "Rule", "evaluate", "Target not match, so the rule evaluates to NotApplicable.");
			result.setDecision(Decision.NOTAPPLICABLE);
			result.setStatus(new Status(StatusCode.STATUS_OK, ""));
			return result;
		}
		// targetΪMatch
		boolean condResult;
		try {
			if(this.condition==null){
				condResult=true;
			}else{
				condResult = this.condition.evaluate(request);
			}
		} catch (IndeterminateException ie) {
			// conditionΪIndeterminateRuleΪIndeterminate
			logger.throwing("Rule", "evaluate", ie);
			result.setDecision(Decision.INDETERMINATE);
			result.setStatus(Status.newStatus(ie));
			return result;
		} 
		if (condResult) {
			if (this.effect == Effect.PERMIT) {
				result.setDecision(Decision.PERMIT);
				result.setStatus(new Status(StatusCode.STATUS_OK, ""));
				return result;
			} else if (this.effect == Effect.DENY) {
				result.setDecision(Decision.DENY);
				result.setStatus(new Status(StatusCode.STATUS_OK, ""));
				return result;
			}
		}
		result.setDecision(Decision.NOTAPPLICABLE);
		result.setStatus(new Status(StatusCode.STATUS_OK, ""));
		return result;
	}

}
